"""
API Command handler module for Bismuth nodes
@EggPoolNet
Needed for Json-RPC server or other third party interaction
"""

import base64
import json
import os
import sys
import threading
from essentials import format_raw_tx

# modular handlers will need access to the database methods under some form, so it needs to be modular too.
# Here, I just duplicated the minimum needed code from node, further refactoring with classes will follow.
import connections
import mempool as mp
from polysign.signerfactory import SignerFactory

__version__ = '0.0.8'


class ApiHandler:
    """
    The API commands manager. Extra commands, not needed for node communication, but for third party tools.
    Handles all commands prefixed by "api_".
    It's called from client threads, so it has to be thread safe.
    """

    __slots__ = ('app_log', 'config', 'callback_lock', 'callbacks')

    def __init__(self, app_log, config=None):
        self.app_log = app_log
        self.config = config
        # Avoid mixing answers to commands with callbacks
        self.callback_lock = threading.Lock()
        # list of sockets that asked for a callback (new block notification)
        # Not used yet.
        self.callbacks = []

    def dispatch(self, method, socket_handler, db_handler, peers):
        """
        Routes the call to the right method
        :return:
        """
        # Easier to ask forgiveness than ask permission
        try:
            """
            All API methods share the same interface. Not storing in properties since it has to be thread safe.
            This is not pretty, this will evolve with more modular code.
            Primary goal is to limit the changes in node.py code and allow more flexibility in this class, like some plugin.
            """
            result = getattr(self, method)(socket_handler, db_handler, peers)
            return result
        except AttributeError:
            # raise
            self.app_log.warning(f'API Method <{method}> does not exist.')
            return False

    def blockstojson(self, raw_blocks: list):
        tx_list = []
        block = {}
        blocks = {}

        old = None
        for transaction_raw in raw_blocks:
            transaction = format_raw_tx(transaction_raw)
            height = transaction['block_height']
            hash = transaction['block_hash']

            del transaction['block_height']
            del transaction['block_hash']

            if old != height:  # if same block
                del tx_list[:]
                block.clear()

            tx_list.append(transaction)

            block['block_height'] = height
            block['block_hash'] = hash
            block['transactions'] = list(tx_list)
            blocks[height] = dict(block)

            old = height  # update

        return blocks

    def blocktojsondiffs(self, list_of_txs: list, list_of_diffs: list):
        i = 0
        blocks_dict = {}
        block_dict = {}
        normal_transactions = []

        old = None
        for transaction in list_of_txs:
            transaction_formatted = format_raw_tx(transaction)
            height = transaction_formatted['block_height']

            del transaction_formatted['block_height']

            #  del transaction_formatted["signature"]  # optional
            #  del transaction_formatted["pubkey"]  # optional

            if old != height:
                block_dict.clear()
                del normal_transactions[:]

            if transaction_formatted['reward'] == 0:  # if normal tx
                del transaction_formatted['block_hash']
                del transaction_formatted['reward']
                normal_transactions.append(transaction_formatted)

            else:
                del transaction_formatted['address']
                del transaction_formatted['amount']
                transaction_formatted['difficulty'] = list_of_diffs[i][0]
                block_dict['mining_tx'] = transaction_formatted

                block_dict['transactions'] = list(normal_transactions)

                blocks_dict[height] = dict(block_dict)
                i += 1
            old = height

        return blocks_dict

    def api_mempool(self, socket_handler, db_handler, peers):
        """
        Returns all the TX from mempool
        :param socket_handler:
        :param db_handler:
        :param peers:
        :return: list of mempool tx
        """
        txs = mp.MEMPOOL.fetchall(mp.SQL_SELECT_TX_TO_SEND)
        connections.send(socket_handler, txs)

    def api_getconfig(self, socket_handler, db_handler, peers):
        """
        Returns configuration
        :param socket_handler:
        :param db_handler:
        :param peers:
        :return: list of node configuration options
        """
        connections.send(socket_handler, self.config.__dict__)

    def api_clearmempool(self, socket_handler, db_handler, peers):
        """
        Empty the current mempool
        :param socket_handler:
        :param db_handler:
        :param peers:
        :return: 'ok'
        """
        mp.MEMPOOL.clear()
        connections.send(socket_handler, 'ok')

    def api_ping(self, socket_handler, db_handler, peers):
        """
        Void, just to allow the client to keep the socket open (avoids timeout)
        :param socket_handler:
        :param db_handler:
        :param peers:
        :return: 'api_pong'
        """
        connections.send(socket_handler, 'api_pong')

    def api_getaddressinfo(self, socket_handler, db_handler, peers):
        """
        Returns a dict with
        known: Did that address appear on a transaction?
        pubkey: The pubkey of the address if it signed a transaction,
        :param address: The bismuth address to examine
        :return: dict
        """
        info = {'known': False, 'pubkey': ''}
        # get the address
        address = connections.receive(socket_handler)
        # print('api_getaddressinfo', address)
        try:
            # format check
            if not SignerFactory.address_is_valid(address):
                self.app_log.info('Bad address format <{}>'.format(address))
                connections.send(socket_handler, info)
                return
            try:
                db_handler.execute_param(db_handler.h, ('SELECT block_height FROM transactions WHERE address= ? or recipient= ? LIMIT 1;'), (address, address))
                _ = db_handler.h.fetchone()[0]
                # no exception? then we have at least one known tx
                info['known'] = True
                db_handler.execute_param(db_handler.h, ('SELECT public_key FROM transactions WHERE address= ? and reward = 0 LIMIT 1;'), (address, ))
                try:
                    info['pubkey'] = db_handler.h.fetchone()[0]
                    info['pubkey'] = base64.b64decode(info['pubkey']).decode('utf-8')
                except Exception as e:
                    self.app_log.warning(e)

            except Exception as e:
                self.app_log.warning(e)

            # returns info
            # print("info", info)
            connections.send(socket_handler, info)
        except Exception as e:
            self.app_log.warning(e)

    def api_getblockfromhash(self, socket_handler, db_handler, peers):
        """
        Returns a specific block based on the provided hash.
        Warning: format is strange: we provide a hash, so there should be at most one result.
        Or we send back a dict, with height as key, and block (including height again) as value.
        Should be enough to only send the block.
        **BUT** do not change, this would break current implementations using the current format (json rpc server for instance).

        :param socket_handler:
        :param db_handler:
        :param peers:
        :return:
        """

        block_hash = connections.receive(socket_handler)

        db_handler.execute_param(
            db_handler.h,
            'SELECT * FROM transactions '
            'WHERE block_hash = ?',
            (block_hash, ),
        )

        result = db_handler.h.fetchall()
        blocks = self.blockstojson(result)
        connections.send(socket_handler, blocks)

    def api_getblockfromhashextra(self, socket_handler, db_handler, peers):
        """
        Returns a specific block based on the provided hash.
        similar to api_getblockfromhash, but sends block dict, not a dict of a dict.
        Also embeds last and next block hash, as well as block difficulty
        Needed for json-rpc server and btc like data.

        :param socket_handler:
        :param db_handler:
        :param peers:
        :return:
        """
        try:
            block_hash = connections.receive(socket_handler)

            result = db_handler.fetchall(
                db_handler.h,
                'SELECT * FROM transactions '
                'WHERE block_hash = ? ',
                (block_hash, ),
            )
            blocks = self.blockstojson(result)
            block = list(blocks.values())[0]

            block['previous_block_hash'] = db_handler.fetchone(db_handler.h, 'SELECT block_hash FROM transactions WHERE block_height = ?', (block['block_height'] - 1, ))
            block['next_block_hash'] = db_handler.fetchone(db_handler.h, 'SELECT block_hash FROM transactions WHERE block_height = ?', (block['block_height'] + 1, ))
            block['difficulty'] = int(float(db_handler.fetchone(db_handler.h, 'SELECT difficulty FROM misc WHERE block_height = ?', (block['block_height'], ))))
            # print(block)
            connections.send(socket_handler, block)
        except Exception as e:
            self.app_log.warning('api_getblockfromhashextra {}'.format(e))
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            self.app_log.warning('{} {} {}'.format(exc_type, fname, exc_tb.tb_lineno))
            raise

    def api_getblockfromheight(self, socket_handler, db_handler, peers):
        """
        Returns a specific block based on the provided hash.

        :param socket_handler:
        :param db_handler:
        :param peers:
        :return:
        """

        height = connections.receive(socket_handler)

        db_handler.execute_param(
            db_handler.h,
            ('SELECT * FROM transactions '
             'WHERE block_height = ? '),
            (height, ),
        )

        result = db_handler.h.fetchall()
        blocks = self.blockstojson(result)
        connections.send(socket_handler, blocks)

    def api_getaddressrange(self, socket_handler, db_handler, peers):
        """
        Returns a given number of transactions, maximum of 500 entries. Ignores blocks where no transactions of a given address happened.
        Reorganizes parameters to a quickly accessible json.
        Unnecessary data are removed.

        :param socket_handler:
        :param db_handler: (UNUSED)
        :param peers: (UNUSED)
        :return:
        """

        address = connections.receive(socket_handler)
        starting_block = connections.receive(socket_handler)
        limit = connections.receive(socket_handler)

        if limit > 500:
            limit = 500

        db_handler.execute_param(db_handler.h, ('SELECT * FROM transactions '
                                                'WHERE ? IN (address, recipient) '
                                                'AND block_height >= ? '
                                                'ORDER BY block_height '
                                                'ASC LIMIT ?'), (
                                                    address,
                                                    starting_block,
                                                    limit,
                                                ))

        result = db_handler.h.fetchall()
        blocks = self.blockstojson(result)
        connections.send(socket_handler, blocks)

    def api_getblockrange(self, socket_handler, db_handler, peers):
        """
        Returns full blocks and transactions from a block range, maximum of 50 entries.
        Includes function format_raw_txs_diffs for formatting. Useful for big data / nosql storage.
        :param socket_handler:
        :param db_handler: (UNUSED)
        :param peers: (UNUSED)
        :return:
        """

        start_block = connections.receive(socket_handler)
        limit = connections.receive(socket_handler)

        if limit > 50:
            limit = 50

        try:
            db_handler.execute_param(db_handler.h, ('SELECT * FROM transactions '
                                                    'WHERE block_height >= ? '
                                                    'AND block_height < ?;'), (
                                                        start_block,
                                                        start_block + limit,
                                                    ))
            raw_txs = db_handler.h.fetchall()

            db_handler.execute_param(db_handler.h, ('SELECT difficulty FROM misc '
                                                    'WHERE block_height >= ? '
                                                    'AND block_height < ?;'), (
                                                        start_block,
                                                        start_block + limit,
                                                    ))
            raw_diffs = db_handler.h.fetchall()

            reply = json.dumps(self.blocktojsondiffs(raw_txs, raw_diffs))

        except Exception as e:
            self.app_log.warning(e)
            raise
        connections.send(socket_handler, reply)

    def api_getblocksince(self, socket_handler, db_handler, peers):
        """
        Returns the full blocks and transactions following a given block_height
        Returns at most transactions from 10 blocks (the most recent ones if it truncates)
        Used by the json-rpc server to poll and be notified of tx and new blocks.

        Returns full blocks and transactions following a given block_height.
        Given block_height should not be lower than the last 10 blocks.
        If given block_height is lower than the most recent block -10,
        last 10 blocks will be returned.

        **Used by the json-rpc server to poll and be notified of tx and new blocks** DO NOT REMOVE!!!.
        :param socket_handler:
        :param db_handler:
        :param peers:
        :return:
        """
        info = []
        # get the last known block
        since_height = connections.receive(socket_handler)
        # print('api_getblocksince', since_height)
        try:
            try:
                db_handler.execute(db_handler.h, 'SELECT MAX(block_height) FROM transactions')
                # what is the min block height to consider ?
                block_height = max(db_handler.h.fetchone()[0] - 11, since_height)
                db_handler.execute_param(db_handler.h, ('SELECT * FROM transactions WHERE block_height > ?;'), (block_height, ))
                info = db_handler.h.fetchall()
                # it's a list of tuples, send as is.
                #print(all)
            except Exception as e:
                print(e)
                raise
            # print("info", info)
            connections.send(socket_handler, info)
        except Exception as e:
            print(e)
            raise

    def api_getblockswhereoflike(self, socket_handler, db_handler, peers):
        """
        Returns the full transactions following a given block_height and with openfield begining by the given string
        Returns at most transactions from 1440 blocks at a time (the most *older* ones if it truncates) so about 1 day worth of data.
        Maybe huge, use with caution and on restrictive queries only.
        :param socket_handler:
        :param db_handler:
        :param peers:
        :return:
        """
        info = []
        # get the last known block
        since_height = int(connections.receive(socket_handler))
        where_openfield_like = connections.receive(socket_handler) + '%'
        #print('api_getblockswhereoflike', since_height, where_openfield_like)
        try:
            try:
                db_handler.execute(db_handler.h, 'SELECT MAX(block_height) FROM transactions')
                # what is the max block height to consider ?
                block_height = min(db_handler.h.fetchone()[0], since_height + 1440)
                #print("block_height", since_height, block_height)
                db_handler.execute_param(db_handler.h, 'SELECT * FROM transactions WHERE block_height > ? and block_height <= ? and openfield like ?', (since_height, block_height, where_openfield_like))
                info = db_handler.h.fetchall()
                # it's a list of tuples, send as is.
                #print("info", info)
            except Exception as e:
                self.app_log.warning(e)
                raise
            # Add the last fetched block so the client will be able to fetch the next block
            info.append([block_height])
            connections.send(socket_handler, info)
        except Exception as e:
            self.app_log.warning(e)
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            self.app_log.warning('{} {} {}'.format(exc_type, fname, exc_tb.tb_lineno))
            raise

    def api_getblocksafterwhere(self, socket_handler, db_handler, peers):
        """
        Returns the full transactions following a given block_height and with specific conditions
        Returns at most transactions from 720 blocks at a time (the most *older* ones if it truncates) so about 12 hours worth of data.
        Maybe huge, use with caution and restrictive queries only.
        :param socket_handler:
        :param db_handler:
        :param peers:
        :return:
        """
        info = []
        # get the last known block
        since_height = connections.receive(socket_handler)
        where_conditions = connections.receive(socket_handler)
        self.app_log.warning('api_getblocksafterwhere', since_height, where_conditions)
        # TODO: feed as array to have a real control and avoid sql injection !important
        # Do *NOT* use in production until it's done.
        raise ValueError('Unsafe, do not use yet')
        """
        [
        ['','openfield','like','egg%']
        ]

        [
        ['', '('],
        ['','reward','>','0']
        ['and','recipient','in',['','','']]
        ['', ')'],
        ]
        """
        where_assembled = where_conditions
        conditions_assembled = ()
        try:
            try:
                db_handler.execute(db_handler.h, 'SELECT MAX(block_height) FROM transactions')
                # what is the max block height to consider ?
                block_height = min(db_handler.h.fetchone()[0], since_height + 720)
                # print("block_height",block_height)
                db_handler.execute_param(db_handler.h, ('SELECT * FROM transactions WHERE block_height > ? and block_height <= ? and ( ' + where_assembled + ')'), (since_height, block_height) + conditions_assembled)
                info = db_handler.h.fetchall()
                # it's a list of tuples, send as is.
                # print(all)
            except Exception as e:
                self.app_log.warning(e)
                raise
            # print("info", info)
            connections.send(socket_handler, info)
        except Exception as e:
            # self.app_log.warning(e)
            raise

    def api_getaddresssince(self, socket_handler, db_handler, peers):
        """
        Returns the full transactions following a given block_height (will not include the given height) for the given address, with at least min_confirmations confirmations,
        as well as last considered block.
        Returns at most transactions from 720 blocks at a time (the most *older* ones if it truncates) so about 12 hours worth of data.

        :param socket_handler:
        :param db_handler:
        :param peers:
        :return:
        """
        info = []
        # get the last known block
        since_height = int(connections.receive(socket_handler))
        min_confirmations = int(connections.receive(socket_handler))
        address = str(connections.receive(socket_handler))
        print('api_getaddresssince', since_height, min_confirmations, address)
        try:
            try:
                db_handler.execute(db_handler.h, 'SELECT MAX(block_height) FROM transactions')
                # what is the max block height to consider ?
                block_height = min(db_handler.h.fetchone()[0] - min_confirmations, since_height + 720)
                db_handler.execute_param(
                    db_handler.h,
                    ('SELECT * FROM transactions WHERE block_height > ? AND block_height <= ? '
                     'AND ((address = ?) OR (recipient = ?)) ORDER BY block_height ASC'),
                    (since_height, block_height, address, address),
                )
                info = db_handler.h.fetchall()
            except Exception as e:
                print('Exception api_getaddresssince:'.format(e))
                raise
            connections.send(socket_handler, {'last': block_height, 'minconf': min_confirmations, 'transactions': info})
        except Exception as e:
            # self.app_log.warning(e)
            raise

    def _get_balance(self, db_handler, address, minconf=1):
        """
        Queries the db to get the balance of a single address
        :param address:
        :param minconf:
        :return:
        """
        try:
            db_handler.execute(db_handler.h, 'SELECT MAX(block_height) FROM transactions')
            # what is the max block height to consider ?
            max_block_height = db_handler.h.fetchone()[0] - minconf
            # calc balance up to this block_height
            db_handler.execute_param(db_handler.h, 'SELECT sum(amount)+sum(reward) FROM transactions WHERE recipient = ? and block_height <= ?;', (address, max_block_height))
            credit = db_handler.h.fetchone()[0]
            if not credit:
                credit = 0
            # debits + fee - reward
            db_handler.execute_param(db_handler.h, 'SELECT sum(amount)+sum(fee) FROM transactions WHERE address = ? and block_height <= ?;', (address, max_block_height))
            debit = db_handler.h.fetchone()[0]
            if not debit:
                debit = 0
            # keep as float
            # balance = '{:.8f}'.format(credit - debit)
            balance = credit - debit
        except Exception as e:
            # self.app_log.warning(e)
            raise
        return balance

    def api_getbalance(self, socket_handler, db_handler, peers):
        """
        returns total balance for a list of addresses and minconf
        BEWARE: this is NOT the json rpc getbalance (that get balance for an account, not an address)
        :param socket_handler:
        :param db_handler:
        :param peers:
        :return:
        """
        balance = 0
        try:
            # get the addresses (it's a list, even if a single address)
            addresses = connections.receive(socket_handler)
            minconf = connections.receive(socket_handler)
            if minconf < 1:
                minconf = 1
            # TODO: Better to use a single sql query with all addresses listed?
            for address in addresses:
                balance += self._get_balance(db_handler, address, minconf)
            # print('api_getbalance', addresses, minconf,':', balance)
            connections.send(socket_handler, balance)
        except Exception as e:
            raise

    def _get_received(self, db_handler, address, minconf=1):
        """
        Queries the db to get the total received amount of a single address
        :param address:
        :param minconf:
        :return:
        """
        try:
            # TODO : for this one and _get_balance, request max block height out of the loop and pass it as a param to alleviate db load
            db_handler.execute(db_handler.h, 'SELECT MAX(block_height) FROM transactions')
            # what is the max block height to consider ?
            max_block_height = db_handler.h.fetchone()[0] - minconf
            # calc received up to this block_height
            db_handler.execute_param(db_handler.h, 'SELECT sum(amount) FROM transactions WHERE recipient = ? and block_height <= ?;', (address, max_block_height))
            credit = db_handler.h.fetchone()[0]
            if not credit:
                credit = 0
        except Exception as e:
            # self.app_log.warning(e)
            raise
        return credit

    def api_getreceived(self, socket_handler, db_handler, peers):
        """
        returns total received amount for a *list* of addresses and minconf
        :param socket_handler:
        :param db_handler:
        :param peers:
        :return:
        """
        received = 0
        try:
            # get the addresses (it's a list, even if a single address)
            addresses = connections.receive(socket_handler)
            minconf = connections.receive(socket_handler)
            if minconf < 1:
                minconf = 1
            # TODO: Better to use a single sql query with all addresses listed?
            for address in addresses:
                received += self._get_received(db_handler, address, minconf)
            print('api_getreceived', addresses, minconf, ':', received)
            connections.send(socket_handler, received)
        except Exception as e:
            # self.app_log.warning(e)
            raise

    def api_listreceived(self, socket_handler, db_handler, peers):
        """
        Returns the total amount received for each given address with minconf, including empty addresses or not.
        :param socket_handler:
        :param db_handler:
        :param peers:
        :return:
        """
        received = {}
        # TODO: this is temporary.
        # Will need more work to send full featured info needed for https://bitcoin.org/en/developer-reference#listreceivedbyaddress
        # (confirmations and tx list)
        try:
            # get the addresses (it's a list, even if a single address)
            addresses = connections.receive(socket_handler)
            minconf = connections.receive(socket_handler)
            if minconf < 1:
                minconf = 1
            include_empty = connections.receive(socket_handler)
            for address in addresses:
                temp = self._get_received(db_handler, address, minconf)
                if include_empty or temp > 0:
                    received[address] = temp
            print('api_listreceived', addresses, minconf, ':', received)
            connections.send(socket_handler, received)
        except Exception as e:
            # self.app_log.warning(e)
            raise

    def api_listbalance(self, socket_handler, db_handler, peers):
        """
        Returns the total amount received for each given address with minconf, including empty addresses or not.
        :param socket_handler:
        :param db_handler:
        :param peers:
        :return:
        """
        balances = {}
        try:
            # get the addresses (it's a list, even if a single address)
            addresses = connections.receive(socket_handler)
            minconf = connections.receive(socket_handler)
            if minconf < 1:
                minconf = 1
            include_empty = connections.receive(socket_handler)
            # TODO: Better to use a single sql query with all addresses listed?
            for address in addresses:
                temp = self._get_balance(db_handler, address, minconf)
                if include_empty or temp > 0:
                    balances[address] = temp
            print('api_listbalance', addresses, minconf, ':', balances)
            connections.send(socket_handler, balances)
        except Exception as e:
            # self.app_log.warning(e)
            raise

    def api_gettransaction(self, socket_handler, db_handler, peers):
        """
        Returns the full transaction matching a tx id. Takes txid anf format as params (json output if format is True)
        :param socket_handler:
        :param db_handler:
        :param peers:
        :return:
        """
        transaction = {}
        try:
            # get the txid
            transaction_id = connections.receive(socket_handler)
            # and format
            format = connections.receive(socket_handler)
            # raw tx details
            if self.config.old_sqlite:
                db_handler.execute_param(db_handler.h, 'SELECT * FROM transactions WHERE signature like ?1', (transaction_id + '%', ))
            else:
                db_handler.execute_param(db_handler.h, 'SELECT * FROM transactions WHERE substr(signature,1,4)=substr(?1,1,4) and signature like ?1', (transaction_id + '%', ))
            raw = db_handler.h.fetchone()
            if not format:
                connections.send(socket_handler, raw)
                print('api_gettransaction', format, raw)
                return

            # current block height, needed for confirmations #
            db_handler.execute(db_handler.h, 'SELECT MAX(block_height) FROM transactions')
            block_height = db_handler.h.fetchone()[0]
            transaction['txid'] = transaction_id
            transaction['time'] = raw[1]
            transaction['hash'] = raw[5]
            transaction['address'] = raw[2]
            transaction['recipient'] = raw[3]
            transaction['amount'] = raw[4]
            transaction['fee'] = raw[8]
            transaction['reward'] = raw[9]
            transaction['operation'] = raw[10]
            transaction['openfield'] = raw[11]
            try:
                transaction['pubkey'] = base64.b64decode(raw[6]).decode('utf-8')
            except:
                transaction['pubkey'] = raw[6]  # support new pubkey schemes
            transaction['blockhash'] = raw[7]
            transaction['blockheight'] = raw[0]
            transaction['confirmations'] = block_height - raw[0]
            # Get more info on the block the tx is in.
            db_handler.execute_param(db_handler.h, 'SELECT timestamp, recipient FROM transactions WHERE block_height= ? AND reward > 0', (raw[0], ))
            block_data = db_handler.h.fetchone()
            transaction['blocktime'] = block_data[0]
            transaction['blockminer'] = block_data[1]
            print('api_gettransaction', format, transaction)
            connections.send(socket_handler, transaction)
        except Exception as e:
            # self.app_log.warning(e)
            raise

    def api_gettransactionbysignature(self, socket_handler, db_handler, peers):
        """
        Returns the full transaction matching a signature. Takes signature and format as params (json output if format is True)
        :param socket_handler:
        :param db_handler:
        :param peers:
        :return:
        """
        transaction = {}
        try:
            # get the txid
            signature = connections.receive(socket_handler)
            # and format
            format = connections.receive(socket_handler)
            # raw tx details
            if self.config.old_sqlite:
                db_handler.execute_param(db_handler.h, 'SELECT * FROM transactions WHERE signature = ?1', (signature, ))
            else:
                db_handler.execute_param(db_handler.h, 'SELECT * FROM transactions WHERE substr(signature,1,4)=substr(?1,1,4) and  signature = ?1', (signature, ))
            raw = db_handler.h.fetchone()
            if not format:
                connections.send(socket_handler, raw)
                print('api_gettransactionbysignature', format, raw)
                return

            # current block height, needed for confirmations
            db_handler.execute(db_handler.h, 'SELECT MAX(block_height) FROM transactions')
            block_height = db_handler.h.fetchone()[0]
            transaction['signature'] = signature
            transaction['time'] = raw[1]
            transaction['hash'] = raw[5]
            transaction['address'] = raw[2]
            transaction['recipient'] = raw[3]
            transaction['amount'] = raw[4]
            transaction['fee'] = raw[8]
            transaction['reward'] = raw[9]
            transaction['operation'] = raw[10]
            transaction['openfield'] = raw[11]
            try:
                transaction['pubkey'] = base64.b64decode(raw[6]).decode('utf-8')
            except:
                transaction['pubkey'] = raw[6]  # support new pubkey schemes
            transaction['blockhash'] = raw[7]
            transaction['blockheight'] = raw[0]
            transaction['confirmations'] = block_height - raw[0]
            # Get more info on the block the tx is in.
            db_handler.execute_param(db_handler.h, 'SELECT timestamp, recipient FROM transactions WHERE block_height= ? AND reward > 0', (raw[0], ))
            block_data = db_handler.h.fetchone()
            transaction['blocktime'] = block_data[0]
            transaction['blockminer'] = block_data[1]
            print('api_gettransactionbysignature', format, transaction)
            connections.send(socket_handler, transaction)
        except Exception as e:
            # self.app_log.warning(e)
            raise

    def api_getpeerinfo(self, socket_handler, db_handler, peers):
        """
        Returns a list of connected peers
        See https://bitcoin.org/en/developer-reference#getpeerinfo
        To be adjusted
        :return: list(dict)
        """
        print('api_getpeerinfo')
        # TODO: Get what we can from peers, more will come when connections and connection stats will be modular, too.
        try:
            info = [{'id': id, 'addr': ip, 'inbound': True} for id, ip in enumerate(peers.consensus)]
            # TODO: peers will keep track of extra info, like port, last time, block_height aso.
            # TODO: add outbound connection
            connections.send(socket_handler, info)
        except Exception as e:
            self.app_log.warning(e)

    def api_gettransaction_for_recipients(self, socket_handler, db_handler, peers):
        """
            Warning: this is currently very slow
            Returns the full transaction matching a tx id for a list of recipient addresses.
            Takes txid and format as params (json output if format is True)
            :param socket_handler:
            :param db_handler:
            :param peers:
            :return:
            """
        transaction = {}
        try:
            # get the txid
            transaction_id = connections.receive(socket_handler)
            # then the recipient list
            addresses = connections.receive(socket_handler)
            # and format
            format = connections.receive(socket_handler)
            recipients = json.dumps(addresses).replace('[', '(').replace(']', ')')  # format as sql
            # raw tx details
            if self.config.old_sqlite:
                db_handler.execute_param(db_handler.h, 'SELECT * FROM transactions WHERE recipient IN {} AND signature LIKE ?1'.format(recipients), (transaction_id + '%', ))
            else:
                db_handler.execute_param(db_handler.h, 'SELECT * FROM transactions WHERE recipient IN {} AND substr(signature,1,4)=substr(?1,1,4) and signature LIKE ?1'.format(recipients), (transaction_id + '%', ))

            raw = db_handler.h.fetchone()
            if not format:
                connections.send(socket_handler, raw)
                print('api_gettransaction_for_recipients', format, raw)
                return

            # current block height, needed for confirmations #
            db_handler.execute(db_handler.h, 'SELECT MAX(block_height) FROM transactions')
            block_height = db_handler.h.fetchone()[0]
            transaction['txid'] = transaction_id
            transaction['time'] = raw[1]
            transaction['hash'] = raw[5]
            transaction['address'] = raw[2]
            transaction['recipient'] = raw[3]
            transaction['amount'] = raw[4]
            transaction['fee'] = raw[8]
            transaction['reward'] = raw[9]
            transaction['operation'] = raw[10]
            transaction['openfield'] = raw[11]

            try:
                transaction['pubkey'] = base64.b64decode(raw[6]).decode('utf-8')
            except:
                transaction['pubkey'] = raw[6]  # support new pubkey schemes

            transaction['blockhash'] = raw[7]
            transaction['blockheight'] = raw[0]
            transaction['confirmations'] = block_height - raw[0]
            # Get more info on the block the tx is in.
            db_handler.execute_param(db_handler.h, 'SELECT timestamp, recipient FROM transactions WHERE block_height= ? AND reward > 0', (raw[0], ))
            block_data = db_handler.h.fetchone()
            transaction['blocktime'] = block_data[0]
            transaction['blockminer'] = block_data[1]
            print('api_gettransaction_for_recipients', format, transaction)
            connections.send(socket_handler, transaction)
        except Exception as e:
            # self.app_log.warning(e)
            raise



#!/usr/bin/env python
# group_access_donor.py
#
# Copyright (C) 2008 Veselin Penev, http://bitdust.io
#
# This file (group_access_donor.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: group_access_donor
.. role:: red

BitDust group_access_donor() Automat

EVENTS:
    * :red:`audit-ok`
    * :red:`fail`
    * :red:`init`
    * :red:`private-key-shared`
    * :red:`shook-hands`
    * :red:`timer-15sec`
    * :red:`timer-30sec`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 6

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.lib import strng

from bitdust.main import events

from bitdust.p2p import handshaker

from bitdust.crypt import my_keys

from bitdust.userid import global_id
from bitdust.userid import id_url

from bitdust.access import key_ring

#------------------------------------------------------------------------------


class GroupAccessDonor(automat.Automat):
    """
    This class implements all the functionality of ``group_access_donor()`` state machine.
    """

    timers = {
        'timer-15sec': (15.0, ['AUDIT']),
        'timer-30sec': (30.0, ['PRIV_KEY']),
    }

    def __init__(self, debug_level=0, log_events=False, log_transitions=False, publish_events=False, **kwargs):
        """
        Builds `group_access_donor()` state machine.
        """
        super(GroupAccessDonor,
              self).__init__(name='group_access_donor', state='AT_STARTUP', debug_level=debug_level or _DebugLevel, log_events=log_events or _Debug, log_transitions=log_transitions or _Debug, publish_events=publish_events, **kwargs)

    def init(self):
        """
        Method to initialize additional variables and flags
        at creation phase of `group_access_donor()` machine.
        """
        self.log_transitions = _Debug
        self.group_key_id = None
        self.result_defer = None

    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python <http://bitdust.io/visio2python/>`_ tool.
        """
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'HANDSHAKE!'
                self.doInit(*args, **kwargs)
                self.doHandshake(*args, **kwargs)
        #---HANDSHAKE!---
        elif self.state == 'HANDSHAKE!':
            if event == 'fail':
                self.state = 'FAILED'
                self.doReportFailed(event, *args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'shook-hands':
                self.state = 'AUDIT'
                self.doAuditUserMasterKey(*args, **kwargs)
        #---AUDIT---
        elif self.state == 'AUDIT':
            if event == 'audit-ok':
                self.state = 'PRIV_KEY'
                self.doSendPrivKeyToUser(*args, **kwargs)
            elif event == 'fail' or event == 'timer-15sec':
                self.state = 'FAILED'
                self.doReportFailed(event, *args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---PRIV_KEY---
        elif self.state == 'PRIV_KEY':
            if event == 'private-key-shared':
                self.state = 'SUCCESS'
                self.doReportDone(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'fail' or event == 'timer-30sec':
                self.state = 'FAILED'
                self.doReportFailed(event, *args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---SUCCESS---
        elif self.state == 'SUCCESS':
            pass
        #---FAILED---
        elif self.state == 'FAILED':
            pass
        return None

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        self.remote_idurl = id_url.field(kwargs['trusted_idurl'])
        self.group_key_id = strng.to_text(kwargs['group_key_id'])
        self.result_defer = kwargs.get('result_defer', None)

    def doHandshake(self, *args, **kwargs):
        """
        Action method.
        """
        d = handshaker.ping(idurl=self.remote_idurl, channel='group_access_donor', keep_alive=True, force_cache=True)
        d.addCallback(lambda ok: self.automat('shook-hands'))
        d.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='group_access_donor.doHandshake')
        d.addErrback(lambda err: self.automat('fail'))

    def doAuditUserMasterKey(self, *args, **kwargs):
        """
        Action method.
        """
        master_key_id = my_keys.make_key_id(alias='master', creator_idurl=self.remote_idurl)
        d = key_ring.audit_private_key(master_key_id, self.remote_idurl)
        d.addCallback(lambda audit_result: (self.automat('audit-ok') if audit_result else self.automat('fail', Exception('remote user master key audit process failed'))))
        d.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='group_access_donor.doAuditUserMasterKey')
        d.addErrback(lambda err: self.automat('fail', err))

    def doSendPrivKeyToUser(self, *args, **kwargs):
        """
        Action method.
        """
        d = key_ring.share_key(self.group_key_id, self.remote_idurl, include_private=True, include_signature=True)
        d.addCallback(self._on_user_priv_key_shared)
        d.addErrback(self._on_user_priv_key_failed)

    def doReportDone(self, *args, **kwargs):
        """
        Action method.
        """
        lg.info('share group key [%s] with %r finished with success' % (self.group_key_id, self.remote_idurl))
        events.send('group-key-shared', data=dict(
            global_id=global_id.UrlToGlobalID(self.remote_idurl),
            remote_idurl=self.remote_idurl,
            group_key_id=self.group_key_id,
        ))
        if self.result_defer:
            self.result_defer.callback(True)

    def doReportFailed(self, event, *args, **kwargs):
        """
        Action method.
        """
        lg.warn('share group key [%s] with %s failed: %s' % (self.group_key_id, self.remote_idurl, args))
        reason = 'group key transfer failed with unknown reason'
        if args and args[0]:
            reason = args[0]
        else:
            if event.count('timer-'):
                reason = 'group key transfer failed because of network connection timeout'
        events.send('group-key-share-failed', data=dict(global_id=global_id.UrlToGlobalID(self.remote_idurl), remote_idurl=self.remote_idurl, group_key_id=self.group_key_id, reason=reason))
        if self.result_defer:
            self.result_defer.errback(Exception(reason))

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        self.destroy()

    def _on_user_priv_key_shared(self, response):
        lg.info('private group key %s was sent to %s' % (self.group_key_id, self.remote_idurl))
        self.automat('private-key-shared', response)
        return None

    def _on_user_priv_key_failed(self, err):
        lg.warn(err)
        self.automat('fail', Exception('private group key delivery failed to remote node'))
        return None



#!/usr/bin/python
# websock.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (websock.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: websock
"""

#------------------------------------------------------------------------------

import os
import time
import json
try:
    from queue import Queue, Empty
except:
    from Queue import Queue, Empty  # @UnresolvedImport

#------------------------------------------------------------------------------

from twisted.internet import reactor  # @UnresolvedImport

#------------------------------------------------------------------------------

from bitdust_forks import websocket

from bitdust.system import local_fs

#------------------------------------------------------------------------------

_Debug = False
_DebugAPIResponses = _Debug

#------------------------------------------------------------------------------

_APISecretFilePath = None
_WebSocketApp = None
_WebSocketQueue = None
_WebSocketReady = False
_WebSocketClosed = True
_WebSocketStarted = False
_WebSocketConnecting = False
_LastCallID = 0
_PendingCalls = []
_CallbacksQueue = {}
_RegisteredCallbacks = {}
_ResponseTimeoutTasks = {}

#------------------------------------------------------------------------------


def start(callbacks={}, api_secret_filepath=None):
    global _APISecretFilePath
    global _WebSocketStarted
    global _WebSocketConnecting
    global _WebSocketQueue
    global _RegisteredCallbacks
    if is_started():
        raise Exception('already started')
    if _Debug:
        print('websock.start()')
    _APISecretFilePath = api_secret_filepath
    _RegisteredCallbacks = callbacks or {}
    _WebSocketConnecting = True
    _WebSocketStarted = True
    _WebSocketQueue = Queue(maxsize=100)
    reactor.callInThread(websocket_thread)  # @UndefinedVariable
    reactor.callInThread(requests_thread, _WebSocketQueue)  # @UndefinedVariable


def stop():
    global _APISecretFilePath
    global _WebSocketStarted
    global _WebSocketQueue
    global _WebSocketConnecting
    global _RegisteredCallbacks
    if not is_started():
        raise Exception('has not been started')
    if _Debug:
        print('websock.stop()')
    _APISecretFilePath = None
    _RegisteredCallbacks = {}
    _WebSocketStarted = False
    _WebSocketConnecting = False
    while True:
        try:
            json_data, _, _, _ = ws_queue().get_nowait()
            if _Debug:
                print('cleaned unfinished call', json_data)
        except Empty:
            break
    _WebSocketQueue.put_nowait((
        None,
        None,
        None,
        None,
    ))
    if ws():
        if _Debug:
            print('websocket already closed')
        ws().close()


#------------------------------------------------------------------------------


def ws():
    global _WebSocketApp
    return _WebSocketApp


def ws_queue():
    global _WebSocketQueue
    return _WebSocketQueue


def is_ready():
    global _WebSocketReady
    return _WebSocketReady


def is_closed():
    global _WebSocketClosed
    return _WebSocketClosed


def is_started():
    global _WebSocketStarted
    return _WebSocketStarted


def is_connecting():
    global _WebSocketConnecting
    return _WebSocketConnecting


def registered_callbacks():
    global _RegisteredCallbacks
    return _RegisteredCallbacks


#------------------------------------------------------------------------------


def on_open(ws_inst):
    global _WebSocketReady
    global _WebSocketClosed
    global _WebSocketConnecting
    global _PendingCalls
    _WebSocketReady = True
    _WebSocketClosed = False
    _WebSocketConnecting = False
    if _Debug:
        print('websocket opened', time.time(), len(_PendingCalls))
    cb = registered_callbacks().get('on_open')
    if cb:
        reactor.callFromThread(cb, ws_inst)  # @UndefinedVariable
    for json_data, cb, tm, timeout in _PendingCalls:
        ws_queue().put_nowait((
            json_data,
            cb,
            tm,
            timeout,
        ))
    _PendingCalls.clear()


def on_close(ws_inst):
    global _WebSocketReady
    global _WebSocketClosed
    global _WebSocketConnecting
    _WebSocketReady = False
    _WebSocketClosed = True
    _WebSocketConnecting = False
    if _Debug:
        print('websocket closed', time.time())
    cb = registered_callbacks().get('on_close')
    if cb:
        reactor.callFromThread(cb, ws_inst)  # @UndefinedVariable


def on_event(json_data):
    if _Debug:
        print('    WS EVENT:', json_data['payload']['event_id'])
    cb = registered_callbacks().get('on_event')
    if cb:
        reactor.callFromThread(cb, json_data)  # @UndefinedVariable
    return True


def on_stream_message(json_data):
    if _Debug:
        print('    WS STREAM MSG:', json_data['payload']['payload']['message_id'])
    cb = registered_callbacks().get('on_stream_message')
    if cb:
        reactor.callFromThread(cb, json_data)  # @UndefinedVariable
    return True


def on_message(ws_inst, message):
    global _CallbacksQueue
    global _ResponseTimeoutTasks
    json_data = json.loads(message)
    if _Debug:
        print('        on_message %d bytes:' % len(message), message)
    if 'payload' not in json_data:
        if _Debug:
            print('        no payload found in the response')
        return False
    payload_type = json_data.get('type')
    if payload_type == 'event':
        return on_event(json_data)
    if payload_type == 'stream_message':
        return on_stream_message(json_data)
    if payload_type == 'api_call':
        if 'call_id' not in json_data['payload']:
            if _Debug:
                print('        call_id not found in the response')
            return
        call_id = json_data['payload']['call_id']
        timeout_task = _ResponseTimeoutTasks.pop(call_id, None)
        if timeout_task:
            if not timeout_task.called:
                timeout_task.cancel()
        if call_id not in _CallbacksQueue:
            if _Debug:
                print('        call_id found in the response, but no callbacks registered')
            return
        result_callback = _CallbacksQueue.pop(call_id, None)
        if _DebugAPIResponses:
            print('WS API Response {} : {}'.format(
                call_id,
                json_data['payload']['response'],
            ))
        if result_callback:
            reactor.callFromThread(result_callback, json_data)  # @UndefinedVariable
        return True
    if _Debug:
        print('        unexpected payload_type', json_data)
    raise Exception(payload_type)


def on_error(ws_inst, error):
    global _PendingCalls
    if _Debug:
        print('on_error', error)
    cb = registered_callbacks().get('on_error')
    if cb:
        reactor.callFromThread(cb, error)  # @UndefinedVariable


def on_fail(err, result_callback=None):
    if _Debug:
        print('on_fail', err)
    if result_callback:
        reactor.callFromThread(result_callback, err)  # @UndefinedVariable


def on_request_timeout(call_id):
    global _CallbacksQueue
    global _ResponseTimeoutTasks
    if _Debug:
        print('on_request_timeout', call_id)
    # timeout_task =
    _ResponseTimeoutTasks.pop(call_id, None)
    # if timeout_task:
    #     if not timeout_task.called:
    #         timeout_task.cancel()
    res_cb = _CallbacksQueue.pop(call_id, None)
    if _DebugAPIResponses:
        print('WS API Request TIMEOUT {}'.format(call_id))
    if res_cb:
        reactor.callFromThread(res_cb, Exception('request timeout'))  # @UndefinedVariable


#------------------------------------------------------------------------------


def requests_thread(active_queue):
    global _LastCallID
    global _CallbacksQueue
    global _ResponseTimeoutTasks
    if _Debug:
        print('\nrequests_thread() starting')
    while True:
        if not is_started():
            if _Debug:
                print('\nrequests_thread() finishing because web socket is not started')
            break
        json_data, result_callback, tm, timeout = active_queue.get()
        if json_data is None:
            if _Debug:
                print('\nrequests_thread() received empty request, about to stop the thread now')
            break
        if 'call_id' not in json_data:
            _LastCallID += 1
            json_data['call_id'] = _LastCallID
        else:
            _LastCallID = json_data['call_id']
        call_id = json_data['call_id']
        if call_id in _CallbacksQueue:
            on_fail(Exception('call_id was not unique'), result_callback)
            continue
        if not ws():
            on_fail(Exception('websocket is closed'), result_callback)
            continue
        _CallbacksQueue[call_id] = result_callback
        data = json.dumps(json_data)
        if _Debug:
            print('sending', data)
        ws().send(data)
        if timeout is not None:
            now = time.time()
            dt = now - tm + timeout
            if dt < 0:
                res_cb = _CallbacksQueue.pop(call_id, None)
                if _DebugAPIResponses:
                    print('\n    WS API Request already TIMED OUT {} : now={} tm={} timeout={}'.format(
                        call_id,
                        now,
                        tm,
                        timeout,
                    ))
                on_fail(Exception('request timeout'), res_cb)
            else:
                _ResponseTimeoutTasks[call_id] = reactor.callLater(dt, on_request_timeout, call_id)  # @UndefinedVariable
    if _Debug:
        print('\nrequests_thread() finished')


def websocket_thread():
    global _APISecretFilePath
    global _WebSocketApp
    global _WebSocketClosed
    websocket.enableTrace(False)
    while is_started():
        _WebSocketClosed = False
        ws_url = 'ws://localhost:8280/'
        if _APISecretFilePath:
            if os.path.isfile(_APISecretFilePath):
                api_secret = local_fs.ReadTextFile(_APISecretFilePath)
                if api_secret:
                    ws_url += '?api_secret=' + api_secret
        if _Debug:
            print('websocket_thread() ws_url=%r' % ws_url)
        _WebSocketApp = websocket.WebSocketApp(
            ws_url,
            on_message=on_message,
            on_error=on_error,
            on_close=on_close,
            on_open=on_open,
        )
        try:
            ws().run_forever(ping_interval=10)
        except Exception as exc:
            _WebSocketApp = None
            if _Debug:
                print('\n    WS Thread ERROR:', exc)
            time.sleep(1)
        if _WebSocketApp:
            del _WebSocketApp
            _WebSocketApp = None
        if not is_started():
            break
        time.sleep(1)
    _WebSocketApp = None


#------------------------------------------------------------------------------


def verify_state():
    global _WebSocketReady
    global _WebSocketConnecting
    if is_closed():
        _WebSocketReady = False
        if _Debug:
            print('WS CALL REFUSED, web socket already closed')
        if is_connecting():
            if _Debug:
                print('web socket closed but still connecting')
            return 'closed'
        return 'closed'
    if is_ready():
        return 'ready'
    if is_connecting():
        return 'connecting'
    if is_started():
        return 'connecting'
    return 'not-started'


#------------------------------------------------------------------------------


def ws_call(json_data, cb=None, timeout=None):
    global _PendingCalls
    st = verify_state()
    if _Debug:
        print('ws_call', st)
    if st == 'ready':
        ws_queue().put_nowait((
            json_data,
            cb,
            time.time(),
            timeout,
        ))
        return True
    if st == 'closed':
        if cb:
            cb(Exception('web socket is closed'))
        return False
    if st == 'connecting':
        if _Debug:
            print('web socket still connecting, remember pending request')
        _PendingCalls.append((
            json_data,
            cb,
            time.time(),
            timeout,
        ))
        return True
    if st == 'not-started':
        if _Debug:
            print('web socket was not started')
        if cb:
            cb(Exception('web socket was not started'))
        return False
    raise Exception('unexpected state %r' % st)



#!/usr/bin/python
# cipher.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (cipher.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
"""
.. module:: cipher.


"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
import base64

try:
    from Cryptodome.Cipher import AES
    from Cryptodome.Cipher import DES3
    from Cryptodome.Util import Padding
    from Cryptodome.Random import get_random_bytes
except:
    from Crypto.Cipher import AES  # @UnresolvedImport @Reimport
    from Crypto.Cipher import DES3  # @UnresolvedImport @Reimport
    from Crypto.Util import Padding  # @UnresolvedImport @Reimport
    from Crypto.Random import get_random_bytes  # @UnresolvedImport @Reimport

#------------------------------------------------------------------------------

from bitdust.lib import serialization

#------------------------------------------------------------------------------


def encrypt_json(raw_data, secret_bytes_key, cipher_type='AES'):
    # TODO: add salt to raw_data
    padded_data = Padding.pad(
        data_to_pad=raw_data,
        block_size=AES.block_size,
    )
    if cipher_type == 'AES':
        cipher = AES.new(
            key=secret_bytes_key,
            mode=AES.MODE_CBC,
        )
    elif cipher_type == 'DES3':
        cipher = DES3.new(
            key=secret_bytes_key,
            mode=DES3.MODE_CBC,
        )
    else:
        raise Exception('unsupported cipher type')
    ct_bytes = cipher.encrypt(padded_data)
    dct = {
        'iv': base64.b64encode(cipher.iv).decode('utf-8'),
        'ct': base64.b64encode(ct_bytes).decode('utf-8'),
    }
    encrypted_data = serialization.DictToBytes(dct, encoding='utf-8')
    return encrypted_data


def decrypt_json(encrypted_data, secret_bytes_key, cipher_type='AES'):
    dct = serialization.BytesToDict(
        encrypted_data,
        encoding='utf-8',
        keys_to_text=True,
        values_to_text=True,
    )
    if cipher_type == 'AES':
        cipher = AES.new(
            key=secret_bytes_key,
            mode=AES.MODE_CBC,
            iv=base64.b64decode(dct['iv'].encode('utf-8')),
        )
    elif cipher_type == 'DES3':
        cipher = DES3.new(
            key=secret_bytes_key,
            mode=DES3.MODE_CBC,
            iv=base64.b64decode(dct['iv'].encode('utf-8')),
        )
    else:
        raise Exception('unsupported cipher type')
    padded_data = cipher.decrypt(base64.b64decode(dct['ct'].encode('utf-8')))
    if cipher_type == 'AES':
        raw_data = Padding.unpad(
            padded_data=padded_data,
            block_size=AES.block_size,
        )
    elif cipher_type == 'DES3':
        raw_data = Padding.unpad(
            padded_data=padded_data,
            block_size=DES3.block_size,
        )
    # TODO: remove salt from raw_data
    return raw_data


#------------------------------------------------------------------------------


def make_key(cipher_type='AES'):
    if cipher_type == 'AES':
        return get_random_bytes(AES.block_size)
    elif cipher_type == 'DES3':
        return get_random_bytes(DES3.block_size)
    raise Exception('unsupported cipher type')


def generate_secret_text(size):
    return base64.b32encode(get_random_bytes(size)).decode()



#!/usr/bin/python
# tcp_node.py
#
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (tcp_node.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: tcp_node.

This is a This is a sub process to send/receive files between users over
TCP protocol. 1) listen for incoming connections on a port 7771 (by
default) 2) establish connections to remote peers 3) keeps TCP session
opened to be able to send asap
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import sys

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in tcp_node.py')

from twisted.internet import protocol  # @UnresolvedImport
from twisted.internet.defer import Deferred  # @UnresolvedImport
from twisted.internet.error import CannotListenError  # @UnresolvedImport

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import net_misc

from bitdust.transport.tcp import tcp_stream

#------------------------------------------------------------------------------

_MyIDURL = None
_MyHost = None
_InternalPort = 7771
_Listener = None
_ByTransferID = {}
_OpenedConnections = {}
_StartedConnections = {}
_ConnectionsCounter = 0
_ConnectionTimeout = 10

#------------------------------------------------------------------------------


def started_connections():
    global _StartedConnections
    return _StartedConnections


def opened_connections():
    global _OpenedConnections
    return _OpenedConnections


def opened_connections_count():
    global _ConnectionsCounter
    return _ConnectionsCounter


#------------------------------------------------------------------------------


def increase_connections_counter():
    global _ConnectionsCounter
    _ConnectionsCounter += 1


def decrease_connections_counter():
    global _ConnectionsCounter
    _ConnectionsCounter -= 1


#------------------------------------------------------------------------------


def get_internal_port():
    global _InternalPort
    return _InternalPort


def my_idurl():
    global _MyIDURL
    return _MyIDURL


def my_host(normalize=False):
    global _MyHost
    if not _MyHost:
        return None
    if normalize:
        return net_misc.normalize_address(_MyHost)
    return _MyHost


#------------------------------------------------------------------------------


def receive(options):
    global _MyIDURL
    global _MyHost
    global _InternalPort
    global _Listener
    from bitdust.transport.tcp import tcp_interface
    if _Debug:
        lg.out(_DebugLevel, 'tcp_node.receive %r' % options)
    if _Listener:
        lg.warn('listener already exist')
        tcp_interface.interface_receiving_started(_MyHost, options)
        return _Listener
    try:
        _MyIDURL = options['idurl']
        _InternalPort = int(options['tcp_port'])
    except:
        _MyIDURL = None
        _InternalPort = None
        lg.exc()
        return None
    try:
        _Listener = reactor.listenTCP(_InternalPort, TCPFactory(None, keep_alive=True))  # @UndefinedVariable
        _MyHost = net_misc.pack_address((options['host'].split(b':')[0], int(_InternalPort)))
        tcp_interface.interface_receiving_started(_MyHost, options)

    except CannotListenError as ex:
        lg.warn('port "%d" is busy' % _InternalPort)
        tcp_interface.interface_receiving_failed('port is busy')
        return None

    except Exception as ex:
        try:
            e = ex.getErrorMessage()
        except:
            e = str(ex)
        tcp_interface.interface_receiving_failed(e)
        lg.exc()
        return None

    return _Listener


def connect_to(host, keep_alive=True):
    host = net_misc.normalize_address(host)
    if host in started_connections():
        lg.warn('already connecting to "%s"' % host)
        return False
    for peeraddr, connections in opened_connections().items():
        if peeraddr == host:
            lg.warn('already connected to "%s" with %d connections' % (host, len(connections)))
            return True
        for connection in connections:
            if connection.getConnectionAddress():
                if connection.getConnectionAddress() == host:
                    lg.warn('already connected to "%s" with peer address: "%s"' % (host, peeraddr))
                    return True
    if _Debug:
        lg.out(_DebugLevel, 'tcp_node.connect_to "%s", keep_alive=%s' % (host, keep_alive))
    connection = TCPFactory(host, keep_alive=keep_alive)
    started_connections()[host] = connection
    connection.connector = reactor.connectTCP(host[0], host[1], connection, timeout=_ConnectionTimeout)  # @UndefinedVariable
    return False


def disconnect_from(host):
    host = net_misc.normalize_address(host)
    ok = False
    for peeraddr, connections in opened_connections().items():
        alll = False
        if peeraddr == host:
            alll = True
        for connection in connections:
            if alll:
                connection.automat('disconnect')
                ok = True
                continue
            if connection.getConnectionAddress():
                if connection.getConnectionAddress() == host:
                    connection.automat('disconnect')
                    return True
    return ok


def disconnect():
    global _Listener
    from bitdust.transport.tcp import tcp_interface
    if not _Listener:
        tcp_interface.interface_disconnected(None)
        return True
    d = _Listener.stopListening()
    if not d:
        tcp_interface.interface_disconnected(None)
    else:
        d.addCallback(tcp_interface.interface_disconnected)
        d.addErrback(lambda *args: lg.warn(str(*args)))
    _Listener = None
    return True


def close_connections():
    for sc in list(started_connections().values()):
        sc.connector.disconnect()
    for oclist in list(opened_connections().values()):
        for oc in oclist:
            oc.automat('disconnect')
            oc.automat('connection-lost')


def send(filename, remoteaddress, description=None, keep_alive=True):
    remoteaddress = net_misc.normalize_address(remoteaddress)
    result_defer = Deferred()
    if remoteaddress == my_host(normalize=True):
        lg.err('sending file %r %r to my own host connection at %r is blocked' % (filename, description, remoteaddress))
        result_defer.callback((filename, description, 'failed', 'cancelled'))
        return result_defer
    if remoteaddress in started_connections():
        started_connections()[remoteaddress].add_outbox_file(filename, description, result_defer, keep_alive)
        if not keep_alive:
            if _Debug:
                lg.out(_DebugLevel, 'tcp_node.send single, use started connection to %s, %d already started and %d opened' % (str(remoteaddress), len(started_connections()), len(opened_connections())))
        return result_defer
    for peeraddr, connections in opened_connections().items():
        for connection in connections:
            if peeraddr == remoteaddress:
                connection.append_outbox_file(filename, description, result_defer, keep_alive)
                if not keep_alive:
                    if _Debug:
                        lg.out(_DebugLevel, 'tcp_node.send single, use opened connection to %s, %d already started and %d opened' % (str(remoteaddress), len(started_connections()), len(opened_connections())))
                return result_defer
            if connection.getConnectionAddress():
                if connection.getConnectionAddress() == remoteaddress:
                    connection.append_outbox_file(filename, description, result_defer, keep_alive)
                    if not keep_alive:
                        if _Debug:
                            lg.out(_DebugLevel, 'tcp_node.send single, use opened connection to %s, %d already started and %d opened' % (str(remoteaddress), len(started_connections()), len(opened_connections())))
                    return result_defer
    if _Debug:
        lg.out(_DebugLevel, 'tcp_node.send start connecting to "%s"' % str(remoteaddress))
    connection = TCPFactory(remoteaddress, keep_alive=keep_alive)
    started_connections()[remoteaddress] = connection
    connection.add_outbox_file(filename, description, result_defer, keep_alive)
    connection.connector = reactor.connectTCP(remoteaddress[0], remoteaddress[1], connection, timeout=_ConnectionTimeout)  # @UndefinedVariable
    if not keep_alive:
        if _Debug:
            lg.out(_DebugLevel, 'tcp_node.send opened a single connection to %s, %d already started and %d opened' % (str(remoteaddress), len(started_connections()), len(opened_connections())))
    return result_defer


def send_keep_alive(host):
    sent_something = False
    for channel in opened_connections().get(host, []):
        channel.automat('send-keep-alive')
        sent_something = True
    return sent_something


def start_streams():
    return tcp_stream.start_process_streams()


def stop_streams():
    return tcp_stream.stop_process_streams()


def list_input_streams(sorted_by_time=True):
    return tcp_stream.list_input_streams(sorted_by_time)


def list_output_streams(sorted_by_time=True):
    return tcp_stream.list_output_streams(sorted_by_time)


def find_stream(file_id=None, transfer_id=None):
    return tcp_stream.find_stream(file_id=file_id, transfer_id=transfer_id)


def cancel_file_receiving(transferID):
    # at the moment for TCP transport we can not stop particular file transfer
    # we can only close connection itself, which is not we really want
    # need to find a way to notify remote side about to stop
    #     for connections in opened_connections().values():
    #         for connection in connections:
    #             for in_file in connection.stream.inboxFiles.values():
    #                 if in_file.transfer_id and in_file.transfer_id == transferID:
    #                     connection.automat('disconnect')
    #                     return True
    #     lg.warn('%r not found' % transferID)
    #     return False
    return False


def cancel_file_sending(transferID):
    for connections in opened_connections().values():
        for connection in connections:
            if connection.stream:
                for out_file in connection.stream.outboxFiles.values():
                    if out_file.transfer_id and out_file.transfer_id == transferID:
                        out_file.cancel()
                        return True
    lg.warn('%r not found' % transferID)
    return False


def cancel_outbox_file(host, filename):
    host = net_misc.normalize_address(host)
    if _Debug:
        lg.args(_DebugLevel, host=host, filename=filename)
    from bitdust.transport.tcp import tcp_interface
    for connections in opened_connections().values():
        for connection in connections:
            if connection.peer_address and connection.peer_address == host or \
                    connection.peer_external_address and connection.peer_external_address == host:
                i = 0
                while i < len(connection.outboxQueue):
                    fn, description, result_defer, keep_alive = connection.outboxQueue[i]
                    if fn == filename:
                        connection.outboxQueue.pop(i)
                        connection.failed_outbox_queue_item(fn, description, 'cancelled')
                        continue
                    i += 1
    for connection in started_connections().values():
        if connection.connection_address and connection.connection_address == host:
            i = 0
            while i < len(connection.pendingoutboxfiles):
                fn, description, result_defer, keep_alive = connection.pendingoutboxfiles[i]
                if fn == filename:
                    connection.pendingoutboxfiles.pop(i)
                    try:
                        tcp_interface.interface_cancelled_file_sending(host, filename, 0, description, 'cancelled')
                    except Exception as exc:
                        lg.warn(str(exc))
                    if result_defer:
                        result_defer.callback((filename, description, 'failed', 'cancelled'))
                    continue
                i += 1


#------------------------------------------------------------------------------


class TCPFactory(protocol.ClientFactory):
    protocol = None

    def __init__(self, connection_address, keep_alive=True):
        from bitdust.transport.tcp import tcp_connection
        self.protocol = tcp_connection.TCPConnection
        self.connection_address = connection_address
        self.keep_alive = keep_alive
        self.pendingoutboxfiles = []
        self.connector = None

    def __repr__(self):
        return 'TCPFactory(%s)' % str(self.connection_address)

    def clientConnectionFailed(self, connector, reason):
        from bitdust.transport.tcp import tcp_interface
        protocol.ClientFactory.clientConnectionFailed(self, connector, reason)
        destaddress = (connector.getDestination().host, int(connector.getDestination().port))
        connection = started_connections().pop(self.connection_address, None)
        if connection:
            connection.connector = None
        if _Debug:
            lg.out(_DebugLevel, 'tcp_node.clientConnectionFailed with %s, %d more connections started : %r' % (str(destaddress), len(started_connections()), reason))
        for filename, description, result_defer, keep_alive in self.pendingoutboxfiles:
            try:
                tcp_interface.interface_cancelled_file_sending(destaddress, filename, 0, description, 'connection failed').addErrback(lambda err: lg.exc(err))
            except Exception as exc:
                lg.warn(str(exc))
            if result_defer:
                result_defer.callback((filename, description, 'failed', 'connection failed'))
        self.pendingoutboxfiles = []

    def add_outbox_file(self, filename, description='', result_defer=None, keep_alive=True):
        self.pendingoutboxfiles.append((filename, description, result_defer, keep_alive))
        tcp_stream.process_streams()



#!/usr/bin/env python
# proxy_router.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (proxy_router.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: proxy_router.

.. role:: red

BitDust proxy_router() Automat

.. raw:: html

    <a href="proxy_router.png" target="_blank">
    <img src="proxy_router.png" style="max-width:100%;">
    </a>

EVENTS:
    * :red:`cancel-route-received`
    * :red:`init`
    * :red:`known-identity-received`
    * :red:`network-connected`
    * :red:`network-disconnected`
    * :red:`request-route-ack-sent`
    * :red:`request-route-received`
    * :red:`routed-inbox-packet-received`
    * :red:`routed-outbox-packet-received`
    * :red:`routed-session-disconnected`
    * :red:`shutdown`
    * :red:`start`
    * :red:`stop`
    * :red:`unknown-identity-received`
    * :red:`unknown-packet-received`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from io import BytesIO

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

_PacketLogFileEnabled = False

#------------------------------------------------------------------------------

import time

#------------------------------------------------------------------------------

from twisted.internet.defer import DeferredList  #@UnresolvedImport
from twisted.internet import reactor  # @UnresolvedImport

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.lib import nameurl
from bitdust.lib import serialization
from bitdust.lib import strng
from bitdust.lib import net_misc

from bitdust.main import config
from bitdust.main import events

from bitdust.services import driver

from bitdust.crypt import key
from bitdust.crypt import signed
from bitdust.crypt import encrypted

from bitdust.userid import identity
from bitdust.userid import my_id
from bitdust.userid import id_url
from bitdust.userid import global_id

from bitdust.contacts import identitydb
from bitdust.contacts import identitycache

from bitdust.transport import callback
from bitdust.transport import packet_out
from bitdust.transport import packet_in
from bitdust.transport import gateway

from bitdust.p2p import p2p_service
from bitdust.p2p import commands
from bitdust.p2p import network_connector

#------------------------------------------------------------------------------

_ProxyRouter = None
_MaxRoutesNumber = 100

#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with proxy_router() machine.
    """
    global _ProxyRouter
    if event is None and not args:
        return _ProxyRouter
    if _ProxyRouter is None:
        # set automat name and starting state here
        _ProxyRouter = ProxyRouter(
            name='proxy_router',
            state='AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=_Debug,
            log_transitions=_Debug,
        )
    if event is not None:
        _ProxyRouter.automat(event, *args, **kwargs)
    return _ProxyRouter


#------------------------------------------------------------------------------


class ProxyRouter(automat.Automat):
    """
    This class implements all the functionality of the ``proxy_router()`` state
    machine.
    """
    def init(self):
        """
        Method to initialize additional variables and flags at creation phase
        of proxy_router() machine.
        """
        self.routes = {}
        self.closed_routes = {}
        self.acks = {}
        self.my_hosts = {}

    def state_changed(self, oldstate, newstate, event, *args, **kwargs):
        """
        Method to catch the moment when proxy_router() state were changed.
        """
        if oldstate != 'TRANSPORTS?' and newstate == 'TRANSPORTS?':
            if network_connector.A().state == 'CONNECTED':
                reactor.callLater(0, self.automat, 'network-connected')  # @UndefinedVariable
            elif network_connector.A().state == 'DISCONNECTED':
                reactor.callLater(0, self.automat, 'network-disconnected')  # @UndefinedVariable

    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python
        <http://code.google.com/p/visio2python/>`_ tool.
        """
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'STOPPED'
                self.doInit(*args, **kwargs)
        #---STOPPED---
        elif self.state == 'STOPPED':
            if event == 'start':
                self.state = 'TRANSPORTS?'
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
        #---LISTEN---
        elif self.state == 'LISTEN':
            if event == 'routed-inbox-packet-received':
                self.doForwardInboxPacket(*args, **kwargs)
                self.doCountIncomingTraffic(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doUnregisterAllRouts(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'routed-outbox-packet-received':
                self.doForwardOutboxPacket(*args, **kwargs)
                self.doCountOutgoingTraffic(*args, **kwargs)
            elif event == 'stop' or event == 'network-disconnected':
                self.state = 'STOPPED'
                self.doUnregisterAllRouts(*args, **kwargs)
            elif event == 'request-route-ack-sent':
                self.doSaveRouteProtoHost(*args, **kwargs)
            elif event == 'known-identity-received':
                self.doSetContactsOverride(*args, **kwargs)
            elif event == 'unknown-identity-received':
                self.doClearContactsOverride(*args, **kwargs)
            elif event == 'unknown-packet-received':
                self.doSendFail(*args, **kwargs)
            elif event == 'request-route-received' or event == 'cancel-route-received':
                self.doProcessRequest(*args, **kwargs)
            elif event == 'routed-session-disconnected':
                self.doUnregisterRoute(*args, **kwargs)
        #---TRANSPORTS?---
        elif self.state == 'TRANSPORTS?':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'stop' or event == 'network-disconnected':
                self.state = 'STOPPED'
            elif event == 'network-connected':
                self.state = 'LISTEN'
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass
        return None

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        global _PacketLogFileEnabled
        _PacketLogFileEnabled = config.conf().getBool('logs/packet-enabled')
        # proxy router must always start with no routes and keep them in memory only
        # when proxy router is restarting all connections with other nodes will be stopped anyway
        if driver.is_on('service_tcp_transport'):
            from bitdust.transport.tcp import tcp_node
            self.my_hosts['tcp'] = tcp_node.my_host(normalize=True)
        if driver.is_on('service_udp_transport'):
            from bitdust.transport.udp import udp_node
            self.my_hosts['udp'] = net_misc.normalize_address(udp_node.A().my_address)
        network_connector.A().addStateChangedCallback(self._on_network_connector_state_changed)
        callback.insert_inbox_callback(0, self._on_first_inbox_packet_received)
        callback.add_finish_file_sending_callback(self._on_finish_file_sending)
        callback.insert_outbox_filter_callback(0, self._on_first_outbox_packet_direct)
        callback.add_file_sending_filter_callback(self._on_file_sending_filter)
        events.add_subscriber(self._on_identity_url_changed, 'identity-url-changed')

    def doProcessRequest(self, *args, **kwargs):
        """
        Action method.
        """
        self._do_process_request(args[0])

    def doUnregisterRoute(self, *args, **kwargs):
        """
        Action method.
        """
        self._do_unregister_route(args[0])

    def doUnregisterAllRouts(self, *args, **kwargs):
        """
        Action method.
        """
        for idurl in list(self.routes.keys()):
            self._do_unregister_route(idurl)
        self.routes.clear()
        self.closed_routes.clear()

    def doForwardOutboxPacket(self, *args, **kwargs):
        """
        Action method.
        """
        self._do_forward_outbox_packet(args[0])

    def doForwardInboxPacket(self, *args, **kwargs):
        """
        Action method.
        """
        self._do_forward_inbox_packet(args[0])

    def doCountOutgoingTraffic(self, *args, **kwargs):
        """
        Action method.
        """

    def doCountIncomingTraffic(self, *args, **kwargs):
        """
        Action method.
        """

    def doSaveRouteProtoHost(self, *args, **kwargs):
        """
        Action method.
        """
        idurl, _, item, _, _, _ = args[0]
        idurl = id_url.field(idurl).original()
        new_address = (
            strng.to_text(item.proto),
            strng.to_text(item.host),
        )
        if idurl not in self.routes:
            lg.exc(exc_value=Exception('route with %r is not registered yet' % idurl))
        else:
            if new_address not in self.routes[idurl]['address']:
                self.routes[idurl]['address'].append(new_address)
                lg.info('added new active address %r for %s, currently %d active addresses' % (new_address, nameurl.GetName(idurl), len(self.routes[idurl]['address'])))

    def doSetContactsOverride(self, *args, **kwargs):
        """
        Action method.
        """
        self._do_set_contacts_override(args[0])

    def doClearContactsOverride(self, *args, **kwargs):
        """
        Action method.
        """
        result = identitycache.StopOverridingIdentity(args[0].CreatorID)
        if _Debug:
            lg.out(_DebugLevel, 'proxy_router.doClearContactsOverride identity for %s, result=%s' % (args[0].CreatorID, result))

    def doSendFail(self, *args, **kwargs):
        """
        Action method.
        """
        newpacket, _ = args[0]
        receiver_idurl = newpacket.OwnerID
        response = 'route not exist'
        if self.closed_routes.get(receiver_idurl.original(), 0) or self.closed_routes.get(receiver_idurl.to_bin(), 0):
            response = 'route already closed'
        if _Debug:
            lg.args(_DebugLevel, response=response, incoming_packet=newpacket)
        p2p_service.SendFail(newpacket, response=response, wide=True)

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        global _PacketLogFileEnabled
        _PacketLogFileEnabled = False
        self.acks.clear()
        for idurl in list(self.routes.keys()):
            self._do_unregister_route(idurl)
        events.remove_subscriber(self._on_identity_url_changed, 'identity-url-changed')
        if network_connector.A():
            network_connector.A().removeStateChangedCallback(self._on_network_connector_state_changed)
        callback.remove_file_sending_filter_callback(self._on_file_sending_filter)
        callback.remove_outbox_filter_callback(self._on_first_outbox_packet_direct)
        callback.remove_inbox_callback(self._on_first_inbox_packet_received)
        callback.remove_finish_file_sending_callback(self._on_finish_file_sending)
        self.my_hosts.clear()
        self.destroy()
        global _ProxyRouter
        del _ProxyRouter
        _ProxyRouter = None

    def _get_session_proto_host(self, sender_idurl, info=None):
        route_info = self.routes.get(sender_idurl.original(), None)
        if not route_info:
            route_info = self.routes.get(sender_idurl.to_bin(), None)
        if not route_info:
            if _Debug:
                lg.dbg(_DebugLevel, 'route with %s was not found' % sender_idurl)
            return None, None
        connection_info = route_info.get('connection_info') or {}
        active_user_session_machine = None
        if (info is not None and not connection_info) or not connection_info.get('index'):
            active_user_sessions = []
            if info:
                active_user_sessions = gateway.find_active_session(info.proto, idurl=sender_idurl.original())
                if not active_user_sessions:
                    active_user_sessions = gateway.find_active_session(info.proto, idurl=sender_idurl.to_bin())
            if not active_user_sessions:
                lg.warn('route with %s found but no active sessions found : %r' % (sender_idurl, info))
                return None, None
            active_user_session_machine = automat.by_index(active_user_sessions[0].index)
        if not active_user_session_machine:
            if connection_info.get('index'):
                active_user_session_machine = automat.by_index(connection_info['index'])
        if not active_user_session_machine:
            lg.warn('route with %s found but no active user session exist' % sender_idurl)
            return None, None
        if not active_user_session_machine.is_connected():
            lg.warn('route with %s found but session is not connected' % sender_idurl)
            return None, None
        hosts = []
        try:
            hosts.append((
                active_user_session_machine.get_proto(),
                active_user_session_machine.get_host(),
            ))
        except:
            lg.exc()
        if not hosts:
            lg.warn('found active user session but host is empty in %r, will try to use recorded info' % active_user_session_machine)
            hosts = route_info['address']
        if len(hosts) == 0:
            lg.warn('route with %s do not have actual info about the host, will use identity contacts instead' % sender_idurl)
            hosts = route_info['contacts']
        if len(hosts) == 0:
            lg.warn('has no known contacts for route with %s' % sender_idurl)
            return None, None
        if len(hosts) > 1:
            lg.warn('found more then one channel with %s : %r' % (sender_idurl, hosts))
        receiver_proto, receiver_host = strng.to_bin(hosts[0][0]), strng.to_bin(hosts[0][1])
        if _Debug:
            lg.args(_DebugLevel, proto=receiver_proto, host=receiver_host, user_session=active_user_session_machine)
        return receiver_proto, receiver_host

    def _do_process_request(self, *args, **kwargs):
        global _MaxRoutesNumber
        json_payload, request, info = args[0]
        user_idurl = request.CreatorID
        #--- commands.RequestService()
        if request.Command == commands.RequestService():
            if len(self.routes) >= _MaxRoutesNumber:
                if _Debug:
                    lg.out(_DebugLevel, 'proxy_server.doProcessRequest RequestService rejected: too many routes')
                p2p_service.SendAck(request, 'rejected', wide=True)
            else:
                try:
                    idsrc = json_payload['identity']
                    cached_ident = identity.identity(xmlsrc=idsrc)
                except:
                    lg.out(_DebugLevel, 'payload: [%s]' % request.Payload)
                    lg.exc()
                    return
                if not cached_ident.Valid():
                    lg.warn('incoming identity is not valid')
                    return
                if not cached_ident.isCorrect():
                    lg.warn('incoming identity is not correct')
                    return
                if user_idurl.original() != cached_ident.getIDURL().original():
                    lg.warn('incoming identity is not belong to request packet creator: %r != %r' % (user_idurl.original(), cached_ident.getIDURL().original()))
                    return
                identitycache.UpdateAfterChecking(cached_ident.getIDURL().original(), idsrc)
                if user_idurl.original() not in list(self.routes.keys()) and user_idurl.to_bin() not in list(self.routes.keys()):
                    oldnew = 'NEW'
                else:
                    oldnew = 'OLD'
                self._do_register_route(user_idurl, cached_ident)
                active_user_sessions = gateway.find_active_session(info.proto, info.host)
                if not active_user_sessions:
                    active_user_sessions = gateway.find_active_session(info.proto, idurl=user_idurl.original())
                if not active_user_sessions:
                    active_user_sessions = gateway.find_active_session(info.proto, idurl=user_idurl.to_bin())
                if active_user_sessions:
                    user_connection_info = {
                        'id': active_user_sessions[0].id,
                        'index': active_user_sessions[0].index,
                        'proto': info.proto,
                        'host': info.host,
                        'idurl': user_idurl,
                    }
                    active_user_session_machine = automat.by_index(user_connection_info['index'])
                    if active_user_session_machine:
                        self.routes[user_idurl.original()]['connection_info'] = user_connection_info
                        active_user_session_machine.addStateChangedCallback(
                            cb=lambda oldstate, newstate, event_string, *args, **kwargs: self._on_user_session_disconnected(user_idurl.original(), oldstate, newstate, event_string, *args, **kwargs),
                            oldstate='CONNECTED',
                            callback_id='proxy_router',
                        )
                        lg.info('connected %s routed user %r and set active session: %r' % (oldnew.upper(), user_idurl, active_user_session_machine))
                    else:
                        lg.err('not found session state machine by index %s' % user_connection_info['index'])
                else:
                    if _Debug:
                        lg.dbg(_DebugLevel, 'active connection with user %s at %s:%s not yet exist' % (user_idurl.original(), info.proto, info.host))
                        lg.dbg(_DebugLevel, 'current active sessions: %d' % len(gateway.list_active_sessions(info.proto)))
                out_ack = p2p_service.SendAck(request, 'accepted', wide=True)
                self.acks[out_ack.PacketID] = out_ack.RemoteID
                if _Debug:
                    lg.out(_DebugLevel, 'proxy_server.doProcessRequest !!!!!!! ACCEPTED %s ROUTE for %r  contacts=%s' % (oldnew.upper(), user_idurl, self.routes.get(user_idurl.original(), {}).get('contacts')))
        #--- commands.CancelService()
        elif request.Command == commands.CancelService():
            if user_idurl.original() in list(self.routes.keys()) or user_idurl.to_bin() in list(self.routes.keys()):
                # cancel existing route
                active_user_session_machine_index = self.routes.get(user_idurl.original(), {}).get('connection_info', {}).get('index', None)
                if active_user_session_machine_index is None:
                    active_user_session_machine_index = self.routes.get(user_idurl.to_bin(), {}).get('connection_info', {}).get('index', None)
                if active_user_session_machine_index is not None:
                    active_user_session_machine = automat.by_index(active_user_session_machine_index)
                    if active_user_session_machine is not None:
                        active_user_session_machine.removeStateChangedCallback(callback_id='proxy_router')
                self.routes.pop(user_idurl.original(), None)
                self.routes.pop(user_idurl.to_bin(), None)
                self.closed_routes[user_idurl.original()] = time.time()
                self.closed_routes[user_idurl.to_bin()] = time.time()
                identitycache.StopOverridingIdentity(user_idurl.original())
                identitycache.StopOverridingIdentity(user_idurl.to_bin())
                p2p_service.SendAck(request, 'accepted', wide=True)
                if _Debug:
                    lg.out(_DebugLevel, 'proxy_server.doProcessRequest !!!!!!! CANCELLED ROUTE for %r' % user_idurl.original())
            else:
                p2p_service.SendAck(request, 'rejected', wide=True)
                if _Debug:
                    lg.out(_DebugLevel, 'proxy_server.doProcessRequest CancelService rejected : %r is not found in routes' % user_idurl.original())
                    lg.out(_DebugLevel, '    %r' % self.routes)
        else:
            p2p_service.SendFail(request, 'rejected', wide=True)

    def _do_forward_inbox_packet(self, *args, **kwargs):
        # encrypt with proxy_receiver()'s key and sent to man behind my proxy
        receiver_idurl, newpacket, info = args[0]
        receiver_idurl = id_url.field(receiver_idurl)
        route_info = self.routes.get(receiver_idurl.original(), None)
        if not route_info:
            route_info = self.routes.get(receiver_idurl.to_bin(), None)
        if _Debug:
            lg.args(_DebugLevel, newpacket=newpacket, info=info, receiver_idurl=receiver_idurl)
        if not route_info:
            lg.warn('route with %s not found for inbox packet: %s' % (receiver_idurl, newpacket))
            return
        connection_info = route_info.get('connection_info', {})
        active_user_session_machine = None
        if not connection_info or not connection_info.get('index'):
            active_user_sessions = gateway.find_active_session(info.proto, idurl=receiver_idurl.original())
            if not active_user_sessions:
                active_user_sessions = gateway.find_active_session(info.proto, idurl=receiver_idurl.to_bin())
            if not active_user_sessions:
                lg.warn('route with %s found but no active sessions found with %s://%s, fire "routed-session-disconnected" event' % (receiver_idurl, info.proto, info.host))
                self.automat('routed-session-disconnected', receiver_idurl)
                return
            user_connection_info = {
                'id': active_user_sessions[0].id,
                'index': active_user_sessions[0].index,
                'proto': info.proto,
                'host': info.host,
                'idurl': receiver_idurl,
            }
            active_user_session_machine = automat.by_index(user_connection_info['index'])
            if active_user_session_machine:
                if receiver_idurl.original() in self.routes:
                    self.routes[receiver_idurl.original()]['connection_info'] = user_connection_info
                    lg.info('found and remember active connection info: %r' % user_connection_info)
                if receiver_idurl.to_bin() in self.routes:
                    self.routes[receiver_idurl.to_bin()]['connection_info'] = user_connection_info
                    lg.info('found and remember active connection info (for latest IDURL): %r' % user_connection_info)
        if not active_user_session_machine:
            if connection_info.get('index'):
                active_user_session_machine = automat.by_index(connection_info['index'])
        if not active_user_session_machine:
            lg.warn('route with %s found but no active user session, fire "routed-session-disconnected" event' % receiver_idurl)
            self.automat('routed-session-disconnected', receiver_idurl)
            return
        if not active_user_session_machine.is_connected():
            lg.warn('route with %s found but session is not connected, fire "routed-session-disconnected" event' % receiver_idurl)
            self.automat('routed-session-disconnected', receiver_idurl)
            return
        hosts = []
        try:
            hosts.append((
                active_user_session_machine.get_proto(),
                active_user_session_machine.get_host(),
            ))
        except:
            lg.exc()
        if not hosts:
            lg.warn('found active user session but host is empty in %r, try use recorded info' % active_user_session_machine)
            hosts = route_info['address']
        if len(hosts) == 0:
            lg.warn('route with %s do not have actual info about the host, use identity contacts instead' % receiver_idurl)
            hosts = route_info['contacts']
        if len(hosts) == 0:
            lg.warn('has no known contacts for route with %s' % receiver_idurl)
            self.automat('routed-session-disconnected', receiver_idurl)
            return
        if len(hosts) > 1:
            lg.warn('found more then one channel with receiver %s : %r' % (receiver_idurl, hosts))
        receiver_proto, receiver_host = strng.to_bin(hosts[0][0]), strng.to_bin(hosts[0][1])
        #--- route is healthy, sending forward incoming routed packet
        raw_data, pout = self._do_send_relay_packet(
            relay_cmd=commands.RelayIn(),
            inbox_packet=newpacket,
            data=newpacket.Serialize(),
            publickey=route_info['publickey'],
            receiver_idurl=receiver_idurl,
            receiver_proto=receiver_proto,
            receiver_host=receiver_host,
            failed_callback=lambda pkt_out, msg: self._on_routed_in_packet_failed(pkt_out, msg, newpacket, info, receiver_idurl),
        )
        if _Debug:
            lg.out(_DebugLevel, '<<<Route-IN %s %s:%s' % (str(newpacket), strng.to_text(info.proto), strng.to_text(info.host)))
            lg.out(_DebugLevel, '           sent to %s://%s with %d bytes in %s' % (strng.to_text(receiver_proto), strng.to_text(receiver_host), len(raw_data), pout))
        active_user_session_machine = None
        del raw_data
        del pout

    def _do_set_contacts_override(self, *args, **kwargs):
        if _Debug:
            lg.out(_DebugLevel, 'proxy_router.doSetContactsOverride identity for %s' % args[0].CreatorID)
        user_idurl = args[0].CreatorID
        idsrc = args[0].Payload
        try:
            new_ident = identity.identity(xmlsrc=idsrc)
        except:
            lg.out(_DebugLevel, 'payload: [%s]' % idsrc)
            lg.exc()
            return
        if not new_ident.isCorrect() or not new_ident.Valid():
            lg.warn('incoming identity is not valid')
            return
        current_overridden_identity = identitycache.ReadOverriddenIdentityXMLSource(user_idurl)
        try:
            current_contacts = identity.identity(xmlsrc=current_overridden_identity).getContacts()
        except:
            current_contacts = []
        identitycache.StopOverridingIdentity(user_idurl)
        result = identitycache.OverrideIdentity(args[0].CreatorID, idsrc)
        if _Debug:
            lg.out(_DebugLevel, '    current overridden contacts is : %s' % current_contacts)
            lg.out(_DebugLevel, '    new override contacts will be : %s' % new_ident.getContacts())
            lg.out(_DebugLevel, '    result=%s' % result)

    def _do_forward_outbox_packet(self, outpacket_info_tuple):
        """
        This packet addressed to me, but contain routed data to be transferred to another node B.
        I will decrypt it with my private key and send the extracted packet to the node B.
        """
        newpacket, info = outpacket_info_tuple
        if _Debug:
            lg.args(_DebugLevel, newpacket=newpacket, info=info)
        block = encrypted.Unserialize(newpacket.Payload)
        if block is None:
            lg.err('failed reading data from %s' % newpacket.RemoteID)
            return
        inpt = None
        try:
            session_key = key.DecryptLocalPrivateKey(block.EncryptedSessionKey)
            padded_data = key.DecryptWithSessionKey(session_key, block.EncryptedData, session_key_type=block.SessionKeyType)
            inpt = BytesIO(padded_data[:int(block.Length)])
            # see proxy_sender.ProxySender : _on_first_outbox_packet() for sending part
            json_payload = serialization.BytesToDict(inpt.read(), keys_to_text=True)
            inpt.close()
            sender_idurl = strng.to_bin(json_payload['f'])  # from
            receiver_idurl = strng.to_bin(json_payload['t'])  # to
            wide = json_payload['w']  # wide
            routed_data = json_payload['p']  # payload
            response_timeout = json_payload.get('i', None)
            keep_alive = json_payload.get('a', False)
            is_retry = json_payload.get('r', False)
        except:
            lg.err('failed reading data from %s' % newpacket.RemoteID)
            lg.exc()
            try:
                inpt.close()
            except:
                pass
            return
        del session_key
        del padded_data
        del inpt
        del block
        if identitycache.HasKey(sender_idurl) and identitycache.HasKey(receiver_idurl) and not is_retry:
            return self._do_verify_routed_data(newpacket, info, sender_idurl, receiver_idurl, routed_data, wide, response_timeout, keep_alive, is_retry)
        lg.warn('will send routed data after caching, is_retry=%s sender_idurl=%r receiver_idurl=%r' % (is_retry, sender_idurl, receiver_idurl))
        dl = []
        if not identitycache.HasKey(sender_idurl) or is_retry:
            dl.append(identitycache.immediatelyCaching(sender_idurl))
        if not identitycache.HasKey(receiver_idurl) or is_retry:
            dl.append(identitycache.immediatelyCaching(receiver_idurl))
        d = DeferredList(dl, consumeErrors=True)
        d.addCallback(self._do_check_cached_idurl, newpacket, info, sender_idurl, receiver_idurl, routed_data, wide, response_timeout, keep_alive, is_retry)
        d.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='_do_forward_outbox_packet')
        d.addErrback(lambda err: self._do_verify_routed_data(newpacket, info, None, None, routed_data, wide, response_timeout, keep_alive, is_retry))
        return True

    def _do_check_cached_idurl(self, cache_results, newpacket, info, sender_idurl, receiver_idurl, routed_data, wide, response_timeout, keep_alive, is_retry):
        sender_id_rev = self.routes.get(sender_idurl, {}).get('identity_rev', None)
        receiver_id_rev = self.routes.get(receiver_idurl, {}).get('identity_rev', None)
        if _Debug:
            lg.args(_DebugLevel, sender_id_rev=sender_id_rev, receiver_id_rev=receiver_id_rev, is_retry=is_retry, cache_results=len(cache_results))
        some_failed = False
        for result, _ in cache_results:
            if not result:
                some_failed = True
        route_changed = False
        if sender_idurl in self.routes:
            sender_ident = identitydb.get_ident(sender_idurl)
            if sender_ident and sender_id_rev is not None and sender_id_rev != sender_ident.getRevisionValue():
                route_changed = True
                self.routes[sender_idurl]['identity_rev'] = sender_ident.getRevisionValue()
                self.closed_routes.pop(sender_idurl, None)
        if receiver_idurl in self.routes:
            receiver_ident = identitydb.get_ident(receiver_idurl)
            if receiver_ident and receiver_id_rev is not None and receiver_id_rev != receiver_ident.getRevisionValue():
                route_changed = True
                self.routes[receiver_idurl]['identity_rev'] = receiver_ident.getRevisionValue()
                self.closed_routes.pop(receiver_idurl, None)
        if some_failed:
            self._do_verify_routed_data(newpacket, info, None, None, routed_data, wide, response_timeout, keep_alive, is_retry, route_changed)
        else:
            self._do_verify_routed_data(newpacket, info, sender_idurl, receiver_idurl, routed_data, wide, response_timeout, keep_alive, is_retry, route_changed)
        return None

    def _do_verify_routed_data(self, newpacket, info, sender_idurl, receiver_idurl, routed_data, wide, response_timeout, keep_alive, is_retry, route_changed=False):
        if sender_idurl is None or receiver_idurl is None:
            lg.warn('failed sending %r, sender or receiver IDURL was not cached' % newpacket)
            self._do_send_fail_packet(newpacket, info, wide, response_timeout, keep_alive, newpacket.CreatorID, receiver_idurl, 'sender or receiver IDURL was not found')
            return
        # those must be already cached
        sender_idurl = id_url.field(sender_idurl)
        receiver_idurl = id_url.field(receiver_idurl)
        route = self.routes.get(sender_idurl.original(), None)
        if not route:
            route = self.routes.get(sender_idurl.to_bin(), None)
        #--- route not exist
        if not route:
            lg.warn('route with %s not exist' % (sender_idurl))
            self._do_send_fail_packet(newpacket, info, wide, response_timeout, keep_alive, sender_idurl, receiver_idurl, 'route not exist')
            return
        routes_keys = list(self.routes.keys())
        closed_route_keys = list(self.closed_routes.keys())
        if _Debug:
            lg.args(_DebugLevel, newpacket=newpacket, info=info, sender_idurl=sender_idurl, receiver_idurl=receiver_idurl, route_contacts=route['contacts'], closed_routes=closed_route_keys, is_retry=is_retry, route_changed=route_changed)
        routed_packet = signed.Unserialize(routed_data)
        #--- invalid packet
        if not routed_packet:
            lg.err('failed to unserialize incoming packet from %s' % newpacket.RemoteID)
            self._do_send_fail_packet(newpacket, info, wide, response_timeout, keep_alive, sender_idurl, receiver_idurl, 'invalid packet')
            return
        routed_command = routed_packet.Command
        routed_packet_id = routed_packet.PacketID
        routed_remote_id = routed_packet.RemoteID
        try:
            is_signature_valid = routed_packet.Valid(raise_signature_invalid=False)
        except:
            is_signature_valid = False
        #--- signature invalid
        if not is_signature_valid:
            lg.err('new packet from %s is NOT VALID:\n\n%r\n\n\n%r\n' % (sender_idurl, routed_data, routed_packet.Serialize()))
            self._do_send_fail_packet(newpacket, info, wide, response_timeout, keep_alive, sender_idurl, receiver_idurl, 'signature invalid')
            return
        #--- packet addressed to me
        if receiver_idurl.to_bin() == my_id.getIDURL().to_bin():
            if _Debug:
                lg.out(_DebugLevel, '        proxy_router() passing by INCOMING packet %r from %s to me' % (routed_packet, sender_idurl))
            # node A sending routed data but I am the actual recipient, so need to handle the packet right away
            packet_in.process(routed_packet, info)
            return
        #--- route already closed
        if False:
            # if receiver_idurl.original() in closed_route_keys or receiver_idurl.to_bin() in closed_route_keys:
            # if not route_changed and ( receiver_idurl.original() in closed_route_keys or receiver_idurl.to_bin() in closed_route_keys ):
            route_closed_time = max(self.closed_routes.get(receiver_idurl.original(), 0), self.closed_routes.get(receiver_idurl.to_bin(), 0))
            if _Debug:
                lg.args(_DebugLevel, route_closed_time=route_closed_time, time=time.time())
            if time.time() - route_closed_time < 900:
                # route was closed but just recently - can not send outgoing data
                lg.err('can not send routed data, route with %s already closed' % (receiver_idurl))
                self._do_send_fail_packet(routed_packet, info, wide, response_timeout, keep_alive, sender_idurl, receiver_idurl, 'route already closed')
                return
        #--- routed inbox packet received
        if receiver_idurl.original() in routes_keys or receiver_idurl.to_bin() in routes_keys:
            # if both node A and node B are behind my router node I need to send routed packet directly to B
            if _Debug:
                lg.out(_DebugLevel, '        proxy_router() ROUTED (same router) packet %s from %s to %s' % (routed_packet, sender_idurl, receiver_idurl))
            self.event('routed-inbox-packet-received', (receiver_idurl, routed_packet, info))
            return
        #--- forward outgoing routed packet
        # send the packet directly to target user
        # do not pass callbacks, because all response packets from this call will be also re-routed
        pout = packet_out.create(
            outpacket=routed_packet,
            wide=wide,
            callbacks={
                'item-failed': lambda pkt_out, out_info: self._on_routed_out_packet_failed(
                    pkt_out,
                    out_info.status,
                    newpacket,
                    info,
                    sender_idurl,
                    routed_command,
                    routed_packet_id,
                    routed_remote_id,
                    wide,
                    response_timeout,
                    keep_alive,
                ),
                'item-sent': lambda pkt_out, out_info: self._on_routed_out_packet_sent(
                    pkt_out,
                    out_info.status,
                    newpacket,
                    info,
                    sender_idurl,
                    routed_command,
                    routed_packet_id,
                    routed_remote_id,
                    wide,
                    response_timeout,
                    keep_alive,
                ),
            },
            target=receiver_idurl,
            response_timeout=response_timeout,
            keep_alive=keep_alive,
            skip_ack=True,
        )
        if _Debug:
            lg.out(_DebugLevel, '>>>Route-OUT %d bytes from %s at %s://%s :' % (len(routed_data), nameurl.GetName(sender_idurl), strng.to_text(info.proto), strng.to_text(info.host)))
            lg.out(_DebugLevel, '    routed to %s : %s' % (nameurl.GetName(receiver_idurl), pout))
        if _PacketLogFileEnabled:
            lg.out(
                0,
                '                \033[0;49;36mROUTE OUT %s(%s) %s %s for %s forwarded to %s\033[0m' % (
                    routed_packet.Command,
                    routed_packet.PacketID,
                    global_id.UrlToGlobalID(routed_packet.OwnerID),
                    global_id.UrlToGlobalID(routed_packet.CreatorID),
                    global_id.UrlToGlobalID(routed_packet.RemoteID),
                    global_id.UrlToGlobalID(receiver_idurl),
                ),
                log_name='packet',
                showtime=True,
            )
        del routed_data
        del route
        del routed_packet

    def _do_send_fail_packet(self, newpacket, info, wide, response_timeout, keep_alive, sender_idurl, receiver_idurl, error):
        if _Debug:
            lg.args(_DebugLevel, error=error, newpacket=newpacket, sender_idurl=sender_idurl, receiver_idurl=receiver_idurl)
        publickey = identitycache.GetPublicKey(newpacket.CreatorID)
        if not publickey:
            lg.err('%r : but can not send RelayFail(), identity %r is not cached' % (error, newpacket.CreatorID))
            return
        receiver_proto, receiver_host = self._get_session_proto_host(sender_idurl, info)
        raw_data, pout = self._do_send_relay_packet(
            relay_cmd=commands.RelayFail(),
            inbox_packet=newpacket,
            data=serialization.DictToBytes(
                {
                    'command': newpacket.Command,
                    'packet_id': newpacket.PacketID,
                    'from': sender_idurl,
                    'to': receiver_idurl,
                    'error': error,
                    'wide': wide,
                    'response_timeout': response_timeout,
                    'keep_alive': keep_alive,
                }
            ),
            publickey=publickey,
            receiver_idurl=sender_idurl,
            receiver_proto=receiver_proto,
            receiver_host=receiver_host,
            error=error,
        )
        if _Debug:
            lg.out(
                _DebugLevel, '<<<Route-FAIL %s from %s:%s   sent to %s://%s with %d bytes in %s' % (
                    str(newpacket),
                    strng.to_text(info.proto),
                    strng.to_text(info.host),
                    strng.to_text(info.proto),
                    strng.to_text(info.host),
                    len(raw_data),
                    pout,
                )
            )
        del raw_data
        del pout

    def _do_send_relay_packet(self, relay_cmd, inbox_packet, data, publickey, receiver_idurl, receiver_proto=None, receiver_host=None, failed_callback=None, error=None):
        if _Debug:
            lg.args(_DebugLevel, relay_cmd=relay_cmd, inbox_packet=inbox_packet, receiver_idurl=receiver_idurl, receiver_proto=receiver_proto, receiver_host=receiver_host)
        block = encrypted.Block(
            CreatorID=my_id.getIDURL(),
            BackupID='routed incoming data',
            BlockNumber=0,
            SessionKey=key.NewSessionKey(session_key_type=key.SessionKeyType()),
            SessionKeyType=key.SessionKeyType(),
            LastBlock=True,
            Data=data,
            EncryptKey=lambda inp: key.EncryptOpenSSHPublicKey(publickey, inp),
        )
        raw_data = block.Serialize()
        routed_packet = signed.Packet(
            Command=relay_cmd,
            OwnerID=inbox_packet.OwnerID,
            CreatorID=my_id.getIDURL(),
            PacketID=inbox_packet.PacketID,
            Payload=raw_data,
            RemoteID=receiver_idurl,
        )
        cbs = {}
        if failed_callback is not None:
            cbs = {
                'failed': failed_callback,
            }
        route = {
            'packet': routed_packet,
            'remoteid': receiver_idurl,
            'description': ('%s_%s[%s]_%s' % (relay_cmd, inbox_packet.Command, inbox_packet.PacketID, nameurl.GetName(receiver_idurl))),
        }
        if receiver_proto is not None and receiver_host is not None:
            route['proto'] = receiver_proto
            route['host'] = receiver_host
        pout = packet_out.create(
            outpacket=inbox_packet,
            wide=False,
            callbacks=cbs,
            route=route,
            skip_ack=True,
        )
        if _PacketLogFileEnabled:
            label = relay_cmd.upper().replace('RELAY', 'ROUTE ')
            reason = error if relay_cmd == commands.RelayFail() else ''
            lg.out(
                0,
                '                \033[0;49;32m%s %s(%s) %s %s for %s forwarded to %s at %s://%s %s\033[0m' % (
                    label,
                    inbox_packet.Command,
                    inbox_packet.PacketID,
                    global_id.UrlToGlobalID(inbox_packet.OwnerID),
                    global_id.UrlToGlobalID(inbox_packet.CreatorID),
                    global_id.UrlToGlobalID(inbox_packet.RemoteID),
                    global_id.UrlToGlobalID(receiver_idurl),
                    strng.to_text(receiver_proto),
                    strng.to_text(receiver_host),
                    reason,
                ),
                log_name='packet',
                showtime=True,
            )
        del block
        del routed_packet
        return raw_data, pout

    def _do_register_route(self, idurl, ident_obj):
        idurl = id_url.field(idurl)
        oldnew = ''
        if idurl.original() not in list(self.routes.keys()) and idurl.to_bin() not in list(self.routes.keys()):
            # accept new route
            oldnew = 'NEW'
            self.routes[idurl.original()] = {}
        else:
            # accept existing routed user
            oldnew = 'OLD'
        if _Debug:
            lg.args(_DebugLevel, idurl=idurl, oldnew=oldnew, rev=ident_obj.getRevisionValue())
        if not self._is_my_contacts_present_in_identity(ident_obj):
            if _Debug:
                lg.out(_DebugLevel, '    DO OVERRIDE identity for %s' % idurl)
            identitycache.OverrideIdentity(idurl, ident_obj.serialize())
        else:
            if _Debug:
                lg.out(_DebugLevel, '        SKIP OVERRIDE identity for %s' % idurl)
        self.routes[idurl.original()]['time'] = time.time()
        self.routes[idurl.original()]['identity'] = ident_obj.serialize(as_text=True)
        self.routes[idurl.original()]['identity_rev'] = ident_obj.getRevisionValue()
        self.routes[idurl.original()]['publickey'] = strng.to_text(ident_obj.publickey)
        self.routes[idurl.original()]['contacts'] = ident_obj.getContactsAsTuples(as_text=True)
        self.routes[idurl.original()]['address'] = []
        self.routes[idurl.original()]['connection_info'] = None
        self.closed_routes.pop(idurl.original(), None)
        self.closed_routes.pop(idurl.to_bin(), None)

    def _do_unregister_route(self, idurl):
        idurl = id_url.field(idurl)
        if _Debug:
            lg.args(_DebugLevel, idurl=idurl)
        active_user_session_machine_index = ((self.routes.get(idurl.original()) or {}).get('connection_info') or {}).get('index', None)
        if active_user_session_machine_index is None:
            active_user_session_machine_index = ((self.routes.get(idurl.to_bin()) or {}).get('connection_info') or {}).get('index', None)
        if active_user_session_machine_index is not None:
            active_user_session_machine = automat.by_index(active_user_session_machine_index)
            if active_user_session_machine is not None:
                active_user_session_machine.removeStateChangedCallback(callback_id='proxy_router')
                lg.info('removed "proxy_router" callback from active user session %r' % active_user_session_machine)
        identitycache.StopOverridingIdentity(idurl.original())
        self.routes.pop(idurl.original(), None)
        self.routes.pop(idurl.to_bin(), None)
        self.closed_routes[idurl.original()] = time.time()
        self.closed_routes[idurl.to_bin()] = time.time()

    def _on_routed_in_packet_failed(self, pkt_out, msg, newpacket, info, receiver_idurl):
        lg.err('routed packet transfer failed: %r %r %r %r %r' % (pkt_out, msg, newpacket, info, receiver_idurl))
        p2p_service.SendFail(newpacket, 'routed packet transfer failed', remote_idurl=newpacket.CreatorID)

    def _on_routed_out_packet_sent(self, pkt_out, msg, newpacket, info, sender_idurl, routed_command, routed_packet_id, routed_remote_id, wide, response_timeout, keep_alive):
        if _Debug:
            lg.args(_DebugLevel, pkt_out=pkt_out, msg=msg, newpacket=newpacket, sender_idurl=sender_idurl, routed_command=routed_command, routed_packet_id=routed_packet_id, routed_remote_id=routed_remote_id)
        publickey = identitycache.GetPublicKey(newpacket.CreatorID)
        if not publickey:
            lg.err('routed packet sent but can not send RelayAck(), identity %r is not cached' % newpacket.CreatorID)
            return
        receiver_proto, receiver_host = self._get_session_proto_host(sender_idurl, info)
        raw_data, pout = self._do_send_relay_packet(
            relay_cmd=commands.RelayAck(),
            inbox_packet=newpacket,
            data=serialization.DictToBytes(
                {
                    'command': routed_command,
                    'packet_id': routed_packet_id,
                    'from': sender_idurl,
                    'to': routed_remote_id,
                    'error': '',
                    'wide': wide,
                    'response_timeout': response_timeout,
                    'keep_alive': keep_alive,
                }
            ),
            publickey=publickey,
            receiver_idurl=sender_idurl,
            receiver_proto=receiver_proto,
            receiver_host=receiver_host,
            error='',
        )
        if _Debug:
            lg.out(_DebugLevel, '<<<Route-ACK %s %s:%s' % (str(newpacket), receiver_proto, receiver_host))
            lg.out(_DebugLevel, '           sent to %s://%s with %d bytes in %s' % (receiver_proto, receiver_host, len(raw_data), pout))
        del raw_data
        del pout
        return None

    def _on_routed_out_packet_failed(self, pkt_out, msg, newpacket, info, sender_idurl, routed_command, routed_packet_id, routed_remote_id, wide, response_timeout, keep_alive):
        if _Debug:
            lg.args(_DebugLevel, pkt_out=pkt_out, msg=msg, newpacket=newpacket, sender_idurl=sender_idurl, routed_command=routed_command, routed_packet_id=routed_packet_id, routed_remote_id=routed_remote_id)
        publickey = identitycache.GetPublicKey(newpacket.CreatorID)
        if not publickey:
            lg.err('routed packet delivery failed but can not send RelayFail(), identity %r is not cached' % newpacket.CreatorID)
            return
        receiver_proto, receiver_host = self._get_session_proto_host(sender_idurl, info)
        raw_data, pout = self._do_send_relay_packet(
            relay_cmd=commands.RelayFail(),
            inbox_packet=newpacket,
            data=serialization.DictToBytes(
                {
                    'command': routed_command,
                    'packet_id': routed_packet_id,
                    'from': sender_idurl,
                    'to': routed_remote_id,
                    'error': 'routed packet delivery failed',
                    'wide': wide,
                    'response_timeout': response_timeout,
                    'keep_alive': keep_alive,
                }
            ),
            publickey=publickey,
            receiver_idurl=sender_idurl,
            receiver_proto=receiver_proto,
            receiver_host=receiver_host,
            error='routed packet delivery failed',
        )
        if _Debug:
            lg.out(_DebugLevel, '<<<Route-FAIL %s %s:%s' % (str(newpacket), receiver_proto, receiver_host))
            lg.out(_DebugLevel, '           sent to %s://%s with %d bytes in %s' % (receiver_proto, receiver_host, len(raw_data), pout))
        del raw_data
        del pout
        return None

    def _on_first_inbox_packet_received(self, newpacket, info, status, error_message):
        if _Debug:
            lg.out(_DebugLevel, 'proxy_router._on_first_inbox_packet_received %s from %s://%s' % (newpacket, info.proto, info.host))
            lg.out(_DebugLevel, '    creator=%s owner=%s' % (newpacket.CreatorID.original(), newpacket.OwnerID.original()))
            lg.out(_DebugLevel, '    sender=%s remote_id=%s' % (info.sender_idurl, newpacket.RemoteID.original()))
            for k, v in self.routes.items():
                lg.out(_DebugLevel*2, '        route with %r :  address=%s  contacts=%s' % (k, v.get('address'), v.get('contacts')))
        # first filter all traffic addressed to me
        if newpacket.RemoteID == my_id.getIDURL():
            # check command type, filter Routed traffic first
            if newpacket.Command in [commands.Relay(), commands.RelayOut()]:
                # look like this is a routed packet from node behind my proxy addressed to someone else
                if (
                    newpacket.CreatorID.original() in list(self.routes.keys()) or newpacket.CreatorID.to_bin() in list(self.routes.keys()) or newpacket.OwnerID.original() in list(self.routes.keys()) or
                    newpacket.OwnerID.to_bin() in list(self.routes.keys())
                ):
                    # sent by proxy_sender() from node A : a man behind proxy_router()
                    # addressed to some third node B in outside world - need to route
                    # A is my consumer and B is a recipient which A wants to contact
                    if _Debug:
                        lg.out(_DebugLevel, '        sending "routed-outbox-packet-received" event')
                    self.event('routed-outbox-packet-received', (newpacket, info))
                    return True
                # looks like we do not know this guy, so why he is sending us routed traffic?
                lg.err('unknown %s from %s received, no known routes with %s' % (newpacket, newpacket.CreatorID, newpacket.CreatorID))
                self.automat('unknown-packet-received', (newpacket, info))
                return True
            # and this is not a Relay packet, Identity
            elif newpacket.Command == commands.Identity():
                # this is a "propagate" packet from node A addressed to this proxy router
                if (newpacket.CreatorID.original() in list(self.routes.keys()) or newpacket.CreatorID.to_bin() in list(self.routes.keys())):
                    # also we need to "reset" overriden identity
                    # return False so that other services also can process that Identity()
                    if _Debug:
                        lg.out(_DebugLevel, '        sending "known-identity-received" event')
                    self.automat('known-identity-received', newpacket)
                    return False
                # this node is not yet in routers list,
                # but seems like it tries to contact me
                # return False so that other services also can process that Identity()
                if _Debug:
                    lg.out(_DebugLevel, '        sending "unknown-identity-received" event')
                self.automat('unknown-identity-received', newpacket)
                return False
            # so this packet may be of any kind, but addressed to me
            # for example if I am a supplier for node A he will send me packets in usual way
            # need to skip this packet here and process it as a normal inbox packet
            if _Debug:
                lg.out(_DebugLevel, '        proxy_router() SKIP packet %s from %s addressed to me' % (newpacket, newpacket.CreatorID))
            return False
        # this packet was addressed to someone else
        # it can be different scenarios, if can not found valid scenario - must skip the packet
        receiver_idurl = None
        known_remote_id = newpacket.RemoteID.original() in list(self.routes.keys()) or newpacket.RemoteID.to_bin() in list(self.routes.keys())
        known_creator_id = newpacket.CreatorID.original() in list(self.routes.keys()) or newpacket.CreatorID.to_bin() in list(self.routes.keys())
        known_owner_id = newpacket.OwnerID.original() in list(self.routes.keys()) or newpacket.OwnerID.to_bin() in list(self.routes.keys())
        if known_remote_id:
            # incoming packet from node B addressed to node A behind that proxy, capture it!
            receiver_idurl = newpacket.RemoteID
            if _Debug:
                lg.out(_DebugLevel, '        proxy_router() ROUTED packet %s from %s to %s' % (newpacket, info.sender_idurl, receiver_idurl))
            self.event('routed-inbox-packet-received', (receiver_idurl, newpacket, info))
            return True
        # unknown RemoteID...
        # Data() packets may have two cases: a new Data or response with existing Data
        # in that case RemoteID of the Data packet is not pointing to the real recipient
        # need to filter this scenario here and do workaround
        if known_creator_id or known_owner_id:
            # response from node B addressed to node A, after Retrieve() from A who owns this Data()
            # a Data packet sent by node B : a man from outside world
            # addressed to a man behind this proxy_router() - need to route to node A
            # but who is node A? Creator or Owner?
            based_on = ''
            if known_creator_id:
                receiver_idurl = newpacket.CreatorID
                based_on = 'creator'
            else:
                receiver_idurl = newpacket.OwnerID
                based_on = 'owner'
            if _Debug:
                lg.out(_DebugLevel, '        proxy_router() based on %s ROUTED packet %s from %s to %s' % (based_on, newpacket, info.sender_idurl, receiver_idurl))
            self.event('routed-inbox-packet-received', (receiver_idurl, newpacket, info))
            return True
        # this packet is not related to any of the routes
        if _Debug:
            lg.dbg(_DebugLevel, 'unknown %r received, no relations found' % newpacket)
        self.automat('unknown-packet-received', (newpacket, info))
        return False

    def _on_first_outbox_packet_direct(self, outpacket, wide, callbacks, target=None, route=None, response_timeout=None, keep_alive=True):
        """
        Will be called first for every outgoing packet.
        When this node A is routing packets for another node B it still must be able to talk to B normally.
        Scenario when packet is routed C -> A -> B is handled in `_on_first_inbox_packet_received()` method.
        Here is another scenario when node A itself wants to talk to B normally - must be no routing in that case.
        The gateway will try to use contacts from the "overridden" identity of the node B - but those "overridden" contacts are
        already pointing to that node A and gateway will try to send a packet to my own host (to host A instead of host B).
        This method filters all packets created by me (not routed via me) and addressed to some of the nodes for whom I am
        already doing proxy routing - those will be re-routed directly to B using the real contacts.
        Must return `None` if that packet should be sent in a normal way - when recipient is not present in my active "routes".
        """
        if self.state != 'LISTEN':
            return None
        if route:
            return None
        if outpacket.CreatorID != my_id.getIDURL():
            return None
        receiver_idurl = outpacket.RemoteID
        receiver_proto, receiver_host = self._get_session_proto_host(receiver_idurl)
        if not receiver_proto or not receiver_host:
            return None
        #--- sending "direct" packet to the node known to be one of my routes
        route = {
            'packet': outpacket,
            'remoteid': receiver_idurl,
            'description': 'direct_%s[%s]_%s' % (outpacket.Command, outpacket.PacketID, nameurl.GetName(receiver_idurl)),
            'proto': receiver_proto,
            'host': receiver_host,
        }
        pkt_out = packet_out.create(
            outpacket=outpacket,
            wide=wide,
            callbacks=callbacks,
            target=target,
            route=route,
            response_timeout=response_timeout,
            keep_alive=keep_alive,
        )
        if _PacketLogFileEnabled:
            lg.out(
                0,
                '\033[1;49;94mOUTBOX DIRECT %s(%s) %s %s to %s at %s://%s\033[0m' % (
                    outpacket.Command,
                    outpacket.PacketID,
                    global_id.UrlToGlobalID(outpacket.OwnerID),
                    global_id.UrlToGlobalID(outpacket.CreatorID),
                    global_id.UrlToGlobalID(receiver_idurl),
                    receiver_proto,
                    receiver_host,
                ),
                log_name='packet',
                showtime=True,
            )
        if _Debug:
            lg.args(_DebugLevel, state=self.state, outpacket=outpacket, wide=wide, route=route, pkt_out=pkt_out)
        return pkt_out

    def _on_file_sending_filter(self, remote_idurl, proto, host, filename, description, pkt_out):
        if id_url.to_bin(remote_idurl) == my_id.getIDURL().to_bin():
            # somehow outgoing file is addressed to my self - do not filter it, but give a warning
            lg.warn('outgoing file addressed to my self: %r' % pkt_out)
            return None
        # now need to check here : the outgoing packet must not be addressed to that host
        # otherwise it must be a "routed" packet - not for me but for another node "routed" via my host
        if proto not in self.my_hosts:
            return None
        if net_misc.normalize_address(host) != self.my_hosts[proto]:
            return None
        remote_idurl = id_url.field(remote_idurl)
        receiver_proto, receiver_host = self._get_session_proto_host(remote_idurl)
        if not receiver_proto or not receiver_host:
            # filter out the packet - because of unknown route we can't send it anyway
            lg.warn('did not found the real host for outgoing %r addressed to my own host' % pkt_out)
            return False
        if _Debug:
            lg.dbg(_DebugLevel, 'switched %s://%s to %s://%s for outgoing %r' % (proto, host, receiver_proto, receiver_host, pkt_out))
        if _PacketLogFileEnabled:
            lg.out(
                0,
                '\033[1;49;94mOUTBOX HOST SWITCH %s://%s to %s://%s for %s towards %s\033[0m' % (
                    proto,
                    host,
                    receiver_proto,
                    receiver_host,
                    description,
                    global_id.UrlToGlobalID(remote_idurl),
                ),
                log_name='packet',
                showtime=True,
            )
        result_defer = gateway.transport(receiver_proto).call('send_file', remote_idurl, filename, receiver_host, description)
        callback.run_begin_file_sending_callbacks(result_defer, remote_idurl, receiver_proto, receiver_host, filename, description, pkt_out)
        # accept the packet and return "filtered" status
        return True

    def _on_network_connector_state_changed(self, oldstate, newstate, event, *args, **kwargs):
        if oldstate != 'CONNECTED' and newstate == 'CONNECTED':
            self.automat('network-connected')
        if oldstate != 'DISCONNECTED' and newstate == 'DISCONNECTED':
            self.automat('network-disconnected')

    def _on_finish_file_sending(self, pkt_out, item, status, size, error_message):
        if status != 'finished':
            return False
        try:
            Command = pkt_out.outpacket.Command
            RemoteID = pkt_out.outpacket.RemoteID
            PacketID = pkt_out.outpacket.PacketID
        except:
            lg.exc()
            return False
        if Command != commands.Ack():
            return False
        if RemoteID.original() not in list(self.routes.keys()) and RemoteID.to_bin() not in list(self.routes.keys()):
            return False
        found = False
        to_remove = []
        for ack_packet_id, ack_remote_idurl in self.acks.items():
            if PacketID.lower() == ack_packet_id.lower() and RemoteID == ack_remote_idurl:
                if _Debug:
                    lg.dbg(_DebugLevel, 'found outgoing Ack() packet %r to %r' % (ack_packet_id, ack_remote_idurl))
                to_remove.append(ack_packet_id)
                # TODO: clean up self.acks for un-acked requests
                self.automat('request-route-ack-sent', (RemoteID, pkt_out, item, status, size, error_message))
                found = True
        for ack_packet_id in to_remove:
            self.acks.pop(ack_packet_id)
        return found

    def _on_user_session_disconnected(self, user_id, oldstate, newstate, event_string, *args, **kwargs):
        lg.warn('user session disconnected  %r : %s->%s' % (user_id, oldstate, newstate))
        self.automat('routed-session-disconnected', user_id)

    def _on_identity_url_changed(self, evt):
        old = id_url.to_bin(evt.data['old_idurl'])
        new = id_url.to_bin(evt.data['new_idurl'])
        if old in self.routes and new not in self.routes:
            current_route = self.routes[old]
            identitycache.StopOverridingIdentity(old)
            self.routes.pop(old)
            self.routes[new] = current_route
            new_ident = identitydb.get_ident(new)
            if new_ident and not self._is_my_contacts_present_in_identity(new_ident):
                if _Debug:
                    lg.out(_DebugLevel, '    DO OVERRIDE identity for %r' % new)
                identitycache.OverrideIdentity(new, new_ident.serialize(as_text=True))
            if new_ident:
                self.routes[new]['identity_rev'] = new_ident.getRevisionValue()
            lg.info('replaced route for user after identity rotate detected : %r -> %r' % (old, new))

    def _is_my_contacts_present_in_identity(self, ident):
        for my_contact in my_id.getLocalIdentity().getContacts():
            if ident.getContactIndex(contact=my_contact) >= 0:
                if _Debug:
                    lg.out(_DebugLevel, '        found %s in identity : %s' % (my_contact, ident.getIDURL()))
                return True
        return False

    def _load_routes(self):
        # TODO: move services/proxy-server/current-routes out from settings into a separate file
        src = config.conf().getString('services/proxy-server/current-routes')
        if src is None:
            lg.warn('setting [services/proxy-server/current-routes] not exist')
            return
        try:
            dct = serialization.BytesToDict(strng.to_bin(src), keys_to_text=True, values_to_text=True)
        except:
            dct = {}
        for k, v in dct.items():
            self.routes[id_url.field(k).original()] = v
            ident = identity.identity(xmlsrc=v['identity'])
            if not self._is_my_contacts_present_in_identity(ident):
                if _Debug:
                    lg.out(_DebugLevel, '    DO OVERRIDE identity for %s' % k)
                identitycache.OverrideIdentity(k, v['identity'])
            else:
                if _Debug:
                    lg.out(_DebugLevel, '        skip overriding %s' % k)
        if _Debug:
            lg.out(_DebugLevel, 'proxy_router._load_routes %d routes total' % len(self.routes))

    def _clear_routes(self):
        # TODO: move services/proxy-server/current-routes out from settings into a separate file
        config.conf().setString('services/proxy-server/current-routes', '{}')
        if _Debug:
            lg.out(_DebugLevel, 'proxy_router._clear_routes')

    def _write_route(self, user_idurl):
        # TODO: move services/proxy-server/current-routes out from settings into a separate file
        src = config.conf().getString('services/proxy-server/current-routes')
        try:
            dct = serialization.BytesToDict(strng.to_bin(src), keys_to_text=True, values_to_text=True)
        except:
            dct = {}
        user_idurl_txt = strng.to_text(id_url.field(user_idurl).original())
        dct[user_idurl_txt] = self.routes[id_url.field(user_idurl).original()]
        newsrc = strng.to_text(serialization.DictToBytes(dct, keys_to_text=True, values_to_text=True))
        config.conf().setString('services/proxy-server/current-routes', newsrc)
        if _Debug:
            lg.out(_DebugLevel, 'proxy_router._write_route %d bytes wrote' % len(newsrc))

    def _remove_route(self, user_idurl):
        # TODO: move services/proxy-server/current-routes out from settings into a separate file
        src = config.conf().getString('services/proxy-server/current-routes')
        try:
            dct = serialization.BytesToDict(strng.to_bin(src), keys_to_text=True, values_to_text=True)
        except:
            dct = {}
        user_idurl_txt = strng.to_text(id_url.field(user_idurl).original())
        if user_idurl_txt in dct:
            dct.pop(user_idurl_txt)
            newsrc = strng.to_text(serialization.DictToBytes(dct, keys_to_text=True, values_to_text=True))
            config.conf().setString('services/proxy-server/current-routes', newsrc)
            if _Debug:
                lg.out(_DebugLevel, 'proxy_router._remove_route %d bytes wrote' % len(newsrc))


#------------------------------------------------------------------------------


def main():
    from twisted.internet import reactor  # @UnresolvedImport
    reactor.callWhenRunning(A, 'init')  # @UndefinedVariable
    reactor.run()  # @UndefinedVariable


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# node.py
#
# Copyright (C) 2007-2008 Francois Aucamp, Meraka Institute, CSIR
# See AUTHORS for all authors and contact information.
#
# License: GNU Lesser General Public License, version 3 or later; see COPYING
#          included in this archive for details.
#
# This library is free software, distributed under the terms of
# the GNU Lesser General Public License Version 3, or any later version.
# See the COPYING file included in this archive
#
# The docstrings in this module contain epytext markup; API documentation
# may be created by processing this file with epydoc: http://epydoc.sf.net

from __future__ import absolute_import
from __future__ import print_function
from io import open
import six

import hashlib

from twisted.internet import defer

from .kademlia.node import Node  # @UnresolvedImport
from .kademlia.node import rpcmethod  # @UnresolvedImport


class EntangledNode(Node):
    """
    Entangled DHT node.

    This is basically a Kademlia node, but with a few more (non-
    standard, but useful) RPCs defined.
    """
    def __init__(self, udpPort=4000, dataStore=None, routingTable=None, networkProtocol=None, **kwargs):
        Node.__init__(self, udpPort, dataStore, routingTable, networkProtocol, **kwargs)
        self.invalidKeywords = []
        self.keywordSplitters = ['_', '.', '/']

    def searchForKeywords(self, keywords):
        """
        The Entangled search operation (keyword-based)

        Call this to find keys in the DHT which contain the specified
        keyword(s).
        """
        if isinstance(keywords, six.string_types):
            for splitter in self.keywordSplitters:
                keywords = keywords.replace(splitter, ' ')
            keywords = keywords.lower().split()

        keyword = None
        for testWord in keywords:
            if testWord not in self.invalidKeywords:
                keyword = testWord
                break
        if keyword is None:
            df = defer.Deferred()
            df.callback([])
            return df

        keywords.remove(keyword)

        h = hashlib.sha1()
        h.update(keyword)
        key = h.digest()

        def checkResult(result):
            if isinstance(result, dict):
                # Value was found; this should be list of "real names" (not keys, in this implementation)
                index = result[key]
                filteredResults = list(index)
                # We found values containing our first keyword; Now filter for the rest
                for name in index:
                    for kw in keywords:
                        if name.lower().find(kw) == -1:
                            filteredResults.remove(name)
                index = filteredResults
            else:
                # Value wasn't found
                index = []
            return index

        df = self.iterativeFindValue(key)
        df.addCallback(checkResult)
        return df

    def publishData(self, name, data):
        """
        The Entangled high-level data publishing operation.

        Call this to store data in the Entangled DHT.

        @note: This will automatically create a hash of the specified C{name}
        parameter, and add the published data to the appropriate inverted
        indexes, to enable keyword-based searching. If this behaviour is not
        wanted/needed, rather call the Kademlia base node's
        C{iterativeStore()} method directly.
        """
        h = hashlib.sha1()
        h.update(name)
        mainKey = h.digest()

        outerDf = defer.Deferred()

        def publishKeywords(deferredResult=None):
            # Create hashes for the keywords in the name
            keywordKeys = self._keywordHashesFromString(name)
            # Update the appropriate inverted indexes
            df = self._addToInvertedIndexes(keywordKeys, name)
            df.addCallback(lambda _: outerDf.callback(None))

        # Store the main key, with its value...
        df = self.iterativeStore(mainKey, data)

        df.addCallback(publishKeywords)

        return outerDf

    def _addToInvertedIndexes(self, keywordKeys, indexLink):
        # Prepare a deferred result for this operation
        outerDf = defer.Deferred()

        kwIndex = [-1]  # using a list for this counter because Python doesn't allow binding a new value to a name in an enclosing (non-global) scope

        # ...and now update the inverted indexes (or add them, if they don't exist yet)
        def addToInvertedIndex(results):
            kwKey = keywordKeys[kwIndex[0]]
            if isinstance(results, dict):
                # An index already exists; add our value to it
                index = results[kwKey]
                # TODO: this might not actually be an index, but a value... do some name-mangling to avoid this
                index.append(indexLink)
            else:
                # An index does not yet exist for this keyword; create one
                index = [indexLink]
            df = self.iterativeStore(kwKey, index)
            df.addCallback(storeNextKeyword)

        def storeNextKeyword(results=None):
            kwIndex[0] += 1
            if kwIndex[0] < len(keywordKeys):
                kwKey = keywordKeys[kwIndex[0]]
                # We use the find algorithm directly so that kademlia does not replicate the un-updated inverted index
                if kwKey in self._dataStore:
                    df = defer.Deferred()
                    df.callback({kwKey: self._dataStore[kwKey]})
                else:
                    df = self._iterativeFind(kwKey, rpc='findValue')
                df.addCallback(addToInvertedIndex)
            else:
                # We're done. Let the caller of the parent method know
                outerDf.callback(None)

        if len(keywordKeys) > 0:
            # Start the "keyword store"-cycle
            storeNextKeyword()
        else:
            outerDf.callback(None)

        return outerDf

    def removeData(self, name):
        """
        The Entangled high-level data removal (delete) operation.

        Call this to remove data from the Entangled DHT.

        @note: This will automatically create a hash of the specified C{name}
        parameter. It will also remove the published data from the appropriate
        inverted indexes, so as to maintain reliability of keyword-based
        searching. If this behaviour is not wanted/needed, rather call this
        node's C{iterativeDelete()} method directly.
        """
        h = hashlib.sha1()
        h.update(name)
        mainKey = h.digest()

        # Remove the main key
        self.iterativeDelete(mainKey)

        # Create hashes for the keywords in the name
        keywordKeys = self._keywordHashesFromString(name)

        # Update the appropriate inverted indexes
        df = self._removeFromInvertedIndexes(keywordKeys, name)
        return df

    def _removeFromInvertedIndexes(self, keywordKeys, indexLink):
        # Prepare a deferred result for this operation
        outerDf = defer.Deferred()

        kwIndex = [-1]  # using a list for this counter because Python doesn't allow binding a new value to a name in an enclosing (non-global) scope

        # ...and now update the inverted indexes (or ignore them, if they don't exist yet)
        def removeFromInvertedIndex(results):
            kwKey = keywordKeys[kwIndex[0]]
            if isinstance(results, dict):
                # An index for this keyword exists; remove our value from it
                index = results[kwKey]
                # TODO: this might not actually be an index, but a value... do some name-mangling to avoid this
                # TODO: this might throw a ValueError; handle it
                index.remove(indexLink)
                # Remove the index completely if it is empty, otherwise put it back
                if len(index) > 0:
                    df = self.iterativeStore(kwKey, index)
                else:
                    df = self.iterativeDelete(kwKey)
                df.addCallback(findNextKeyword)
            else:
                # No index exists for this keyword; skip it
                findNextKeyword()

        def findNextKeyword(results=None):
            kwIndex[0] += 1
            if kwIndex[0] < len(keywordKeys):
                kwKey = keywordKeys[kwIndex[0]]
                # We use the find algorithm directly so that kademlia does not replicate the un-updated inverted index
                if kwKey in self._dataStore:
                    df = defer.Deferred()
                    df.callback({kwKey: self._dataStore[kwKey]})
                else:
                    df = self._iterativeFind(kwKey, rpc='findValue')
                df.addCallback(removeFromInvertedIndex)
            else:
                # We're done. Let the caller of the parent method know
                outerDf.callback(None)

        if len(keywordKeys) > 0:
            # Start the "keyword store"-cycle
            findNextKeyword()

        return outerDf

    def iterativeDelete(self, key):
        """
        The Entangled delete operation.

        Call this to remove data from the DHT.

        The Entangled delete operation uses the basic Kademlia node lookup
        algorithm (same as Kademlia's search/retrieve). The algorithm behaves
        the same as when issueing the FIND_NODE RPC - the only difference is
        that the DELETE RPC (defined in C{delete()}) is used instead of
        FIND_NODE.

        @param key: The hashtable key of the data
        @type key: str
        """
        # Delete our own copy of the data
        if key in self._dataStore:
            del self._dataStore[key]
        df = self._iterativeFind(key, rpc='delete')
        return df

    @rpcmethod
    def delete(self, key, **kwargs):
        """
        Deletes the the specified key (and it's value) if present in this
        node's data, and executes FIND_NODE for the key.

        @param key: The hashtable key of the data to delete
        @type key: str

        @return: A list of contact triples closest to the specified key.
                 This method will return C{k} (or C{count}, if specified)
                 contacts if at all possible; it will only return fewer if the
                 node is returning all of the contacts that it knows of.
        @rtype: list
        """
        # Delete our own copy of the data (if we have one)...
        if key in self._dataStore:
            del self._dataStore[key]
        # ...and make this RPC propagate through the network (like a FIND_VALUE for a non-existant value)
        return self.findNode(key, **kwargs)

    def _keywordHashesFromString(self, text):
        """
        Create hash keys for the keywords contained in the specified text
        string.
        """
        keywordKeys = []
        splitText = text.lower()
        for splitter in self.keywordSplitters:
            splitText = splitText.replace(splitter, ' ')
        for keyword in splitText.split():
            # Only consider keywords with 3 or more letters
            if len(keyword) >= 3 and keyword != text and keyword not in self.invalidKeywords:
                h = hashlib.sha1()
                h.update(keyword)
                key = h.digest()
                keywordKeys.append(key)
        return keywordKeys


if __name__ == '__main__':
    import twisted.internet.reactor
    from kademlia.datastore import SQLiteVersionedJsonDataStore  # @UnresolvedImport
    import sys
    if len(sys.argv) < 2:
        print('Usage:\n%s UDP_PORT  [KNOWN_NODE_IP  KNOWN_NODE_PORT]' % sys.argv[0])
        print('or:\n%s UDP_PORT  [FILE_WITH_KNOWN_NODES]' % sys.argv[0])
        print('\nIf a file is specified, it should containg one IP address and UDP port\nper line, seperated by a space.')
        sys.exit(1)
    try:
        int(sys.argv[1])
    except ValueError:
        print('\nUDP_PORT must be an integer value.\n')
        print('Usage:\n%s UDP_PORT  [KNOWN_NODE_IP  KNOWN_NODE_PORT]' % sys.argv[0])
        print('or:\n%s UDP_PORT  [FILE_WITH_KNOWN_NODES]' % sys.argv[0])
        print('\nIf a file is specified, it should contain one IP address and UDP port\nper line, seperated by a space.')
        sys.exit(1)

    if len(sys.argv) == 4:
        knownNodes = [(sys.argv[2], int(sys.argv[3]))]
    elif len(sys.argv) == 3:
        knownNodes = []
        f = open(sys.argv[2], 'r')
        lines = f.readlines()
        f.close()
        for line in lines:
            ipAddress, udpPort = line.split()
            knownNodes.append((ipAddress, int(udpPort)))
    else:
        knownNodes = None

    # if os.path.isfile('/tmp/dbFile%s.db' % sys.argv[1]):
    #     os.remove('/tmp/dbFile%s.db' % sys.argv[1])
    dataStore = SQLiteVersionedJsonDataStore(dbFile='/tmp/dbFile%s.db' % sys.argv[1])
    node = EntangledNode(udpPort=int(sys.argv[1]), dataStore=dataStore)
    #node = EntangledNode( udpPort=int(sys.argv[1]) )
    node.joinNetwork(knownNodes)
    twisted.internet.reactor.run()  # @UndefinedVariable



#!/usr/bin/python
# service_network.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_network.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_network
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return NetworkService()


class NetworkService(LocalService):

    service_name = 'service_network'
    config_path = 'services/network/enabled'

    current_network_interfaces = None

    def dependent_on(self):
        return [
            # this is a top root service, everything in BitDust depends on networking
        ]

    def start(self):
        from twisted.internet import task
        from bitdust.main import events
        from bitdust.p2p import network_connector
        network_connector.A('init')
        self.task = task.LoopingCall(self._do_check_network_interfaces)
        self.task.start(20, now=False)
        events.add_subscriber(self._on_my_identity_rotate_complete, 'my-identity-rotate-complete')
        events.add_subscriber(self._on_my_external_ip_changed, 'my-external-ip-changed')
        return True

    def stop(self):
        from bitdust.main import events
        from bitdust.p2p import network_connector
        events.remove_subscriber(self._on_my_external_ip_changed, 'my-external-ip-changed')
        events.remove_subscriber(self._on_my_identity_rotate_complete, 'my-identity-rotate-complete')
        network_connector.Destroy()
        if self.task and self.task.running:
            self.task.stop()
            self.task = None
        return True

    def _on_my_external_ip_changed(self, evt):
        from bitdust.logs import lg
        if evt.data['old'].strip():
            lg.info('need to reconnect because my external IP changed %r -> %r' % (evt.data['old'], evt.data['new']))
            from bitdust.p2p import network_connector
            network_connector.A('reconnect')

    def _on_my_identity_rotate_complete(self, evt):
        from bitdust.logs import lg
        from bitdust.services import driver
        if driver.is_enabled('service_gateway'):
            lg.warn('my identity sources were rotated, need to restart service_gateway()')
            #             if driver.is_enabled('service_identity_propagate'):
            #                 from bitdust.p2p import propagate
            #                 from bitdust.contacts import contactsdb
            #                 selected_contacts = set(filter(None, contactsdb.contacts_remote(include_all=True)))
            #                 if propagate.startup_list():
            #                     selected_contacts.update(propagate.startup_list())
            #                     propagate.startup_list().clear()
            #                 propagate.propagate(
            #                     selected_contacts=selected_contacts,
            #                     wide=True,
            #                     refresh_cache=True,
            #                 ).addBoth(lambda err: driver.restart('service_gateway'))
            #             else:
            driver.restart('service_gateway')
        else:
            lg.warn('my identity sources were rotated, but service_gateway() is disabled')
        return None

    def _do_check_network_interfaces(self):
        from bitdust.lib.net_misc import getNetworkInterfaces
        from bitdust.p2p import network_connector
        from bitdust.logs import lg
        known_interfaces = getNetworkInterfaces()
        if '127.0.0.1' in known_interfaces:
            known_interfaces.remove('127.0.0.1')
        if self.current_network_interfaces is None:
            self.current_network_interfaces = known_interfaces
            lg.info('current network interfaces on START UP: %r' % self.current_network_interfaces)
        else:
            if self.current_network_interfaces != known_interfaces and known_interfaces:
                lg.info('need to reconnect, recognized changes in network interfaces: %r -> %r' % (self.current_network_interfaces, known_interfaces))
                self.current_network_interfaces = known_interfaces
                network_connector.A('reconnect')
